'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''

import io
import os
import sys
import gzip
import hashlib
import yaml
import json
import time
from datetime import datetime
from math import ceil
from itertools import chain, repeat


from Bio import SeqIO


# from sqlalchemy.orm import sessionmaker
# from sqlalchemy.orm.attributes import flag_modified
# from sqlalchemy.ext.declarative import declarative_base
# from sqlalchemy import Table, Column, Integer, String, MetaData, ForeignKey, Sequence, JSON, Boolean

#from psycopg2 import sql
import tempfile
import numpy as np


from kmerdb import kmer





def parsefile(filepath:str, k:int, logger=None): #rows_per_batch:int=100000, b:int=50000, n:int=1000, both_strands:bool=False, all_metadata:bool=False):
    """Parse a single sequence file in blocks/chunks with multiprocessing support

    :param filepath: Path to a fasta or fastq file
    :type filepath: str
    :param k: Choice of k to shred k-mers with
    :type k: int
    :param b: Number of reads (per block) to process in parallel
    :type b: int
    :param stranded: Strand specificity argument for k-mer shredding process
    :type stranded: bool
    :raise TypeError: filepath was invalid
    :raise OSError: filepath was invalid
    :raise TypeError: k was invalid
    :raise TypeError: invalid dtype
    :raise ValueError: invalid (None) kmer id detected
    :raise ValueError: mismatched number of kmer_ids, associated sequence/read ids, starting locations, and reverse bools
    :raise AssertionError: Error in nullomer count estimation
    :returns: (counts, header_dictionary, nullomer_array, all_metadata) header_dictionary is the file's metadata for the header block
    :rtype: (numpy.ndarray, dict, list, list)

    """

    from kmerdb import kmer
    if filepath is None or type(filepath) is not str:
        raise TypeError("kmerdb.parse.parsefile expects a str as its first positional argument")
    elif not os.path.exists(filepath):
        raise OSError("kmerdb.parse.parsefile could not find the file '{0}' on the filesystem".format(filepath))
    elif k is None or type(k) is not int:
        raise TypeError("kmerdb.parse.parsefile expects an int as its second positional argument")
    # elif rows_per_batch is None or type(rows_per_batch) is not int:
    #     raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'rows_per_batch' to be an int")
    # elif b is None or type(b) is not int:
    #     raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'b' to be an int")
    # elif n is None or type(n) is not int:
    #     raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'n' to be an int")
    # elif both_strands is None or type(both_strands) is not bool:
    #     raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'both_strands' to be a bool")
    # elif all_metadata is None or type(all_metadata) is not bool:
    #     raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'all_metadata' to be a bool")
    # if logger is None:
    #     raise ValueError("Invalid logger")

    _loggable = logger is not None
    
    data = {} # This is the dictionary of tuples, keyed on k-mer id, and containing 3-tuples ('kmer_id', 'read/start/reverse')
    keys = set()
    rows = []
    N = 4**k
    nullomers = set()

    
    # Build fasta/fastq parser object to stream reads into memory
    if _loggable:
        logger.log_it("Initializing parser...", "INFO")


    
    seqprsr = SeqParser(filepath, k, logger)
    fasta = not seqprsr.fastq # Look inside the seqprsr object for the type of file

    # Initialize the kmer array
    try:
        counts = np.zeros(N, dtype="uint64")
    except TypeError as e:
        if _loggable:
            logger.log_it("Invalid dtype for numpy array instantiation", "ERROR")
            logger.log_it(e.__str__(), "ERROR")
        raise e

    if _loggable:
        logger.log_it("Successfully allocated space for {0} unsigned integers: {1} bytes".format(N, counts.nbytes), "INFO")
        # Instantiate the kmer class
    Kmer = kmer.Kmers(k, verbose=fasta) # A wrapper class to shred k-mers with

    recs = [r for r in seqprsr] # A block of exactly 'b' reads-per-block to process in parallel
    if not fasta:
        if _loggable:
            logger.log_it("Read exactly {0} records from the seqparser object".format(len(recs)), "DEBUG")
        assert len(recs) <= b, "The seqparser should return exactly {0} records at a time".format(b)
    elif _loggable:
        logger.log_it("Skipping the block size assertion for fasta files", "DEBUG")

    if _loggable:
        logger.log_it("Read {0} sequences from the {1} seqparser object".format(len(recs), "fasta" if fasta else "fastq"), "INFO")
    all_kmer_metadata = list([] for x in range(N))
    while len(recs): # While the seqprsr continues to produce blocks of reads

        num_recs = len(recs)

        if _loggable:
            logger.log_it("Processing a block of {0} reads/sequences".format(num_recs), "INFO")

        # Initialize an all_kmer_metadata list
        kmer_metadata = []
        list_of_dicts = list(map(Kmer.shred, recs))

        # Flatmap to 'kmers', the dictionary of {'id': read_id, 'kmers': [ ... ]}
        if _loggable:
            logger.log_it("\n\nAcquiring linked-list of all k-mer ids from {0} sequence records...\n\n".format(num_recs), "INFO")
        kmer_ids = list(chain.from_iterable(map(lambda x: x['kmers'], list_of_dicts)))

        sus = set()

        
        if _loggable:
            logger.log_it("Removing any eroneous k-mers without an id, cause by 'N' content, creating new gaps in the k-mer profile...", "WARNING")
            logger.log_it("Will substitute IUPAC residues with their respective residues, adding one count for each", "WARNING")
        num_sus = 0
        for i, k in enumerate(kmer_ids):
            if k is None:
                sus.add(i) # This propagates the N removal, by adding those to the set for removal later.
                num_sus += 1

        logger.log_it("Created {0} gaps in the k-mer profile to clean it of N-content".format(num_sus), "INFO")
        if num_sus > 0 and _loggable:
            logger.log_it("{0} uncountable k-mer identified".format(num_sus), "WARNING")
            logger.log_it("Likely caused by the presence of the unspecified nucleotide 'N' in a .fasta file", "WARNING")
            logger.log_t("Will remove from the final committed profile", "WARNING")

        ########################
        # K-mer cleaning
        #
        # 
        # The following procedure, removes k-mers that don't have a formal k-mer id
        # More specifically, it eliminates entries that were formally loaded into the k-mers array,
        # and thus would be a minor percentage of the total number of k-mers,
        # Specifically, k-mers or subsequences with N can be excluded from the profile
        # To essentially eliminate an area of sequence and create an artificial gap in the "retained" graph.
        # The graph can be explicitly collapsed when you want to, by supplying reads or sequence information, that is contiguous "in your judgement".
        # By introducing these artificial gaps, we expose the errors or unspecified bases through data.
        #

        if _loggable:
            logger.log_it("Found {0} invalid or enumerable k-mers, sequences which have no defined index, i".format(num_sus), "INFO")
            logger.log_it("In other words, one or more characters in your .fasta or .fastq file were 'N' or otherwise contained IUPAC characters that need to be substituted.", "DEBUG")
            logger.log_it("We have chosen to omit k-mer with unspecified nucleotides 'N', and these make gaps in our database explicitly.", "DEBUG")
            logger.log_it("These can be recovered from the reads if they are needed, and the read ids may be found with the experimental --all-metadata flag", "INFO")
        # if all_metadata:

        #     logger.warning("\n\n\nGENERATING ALL METADATA\n\n\nThis is extremely expensive and experimental. You have been warned.\n\n\n")
        #     reads = list(chain.from_iterable(map(lambda x: x['seqids'], list_of_dicts)))
        #     starts = list(chain.from_iterable(map(lambda x: x['starts'], list_of_dicts)))
        #     reverses = list(chain.from_iterable(map(lambda x: x['reverses'], list_of_dicts)))
        #     logger.debug("Checking each k-mer for N-content and IUPAC substitution")
        #     for i, x in enumerate(kmer_ids): # This removes N content
        #         if i in sus: # This is where we actually delete the N content, in case that is eventually supported.
        #             kmer_ids[i] = None
        #             reads[i] = None
        #             starts[i] = None
        #             reverses[i] = None
        #     kmer_ids = list(filter(lambda k: k is not None, kmer_ids)) 
        #     reads = list(filter(lambda r: r is not None, reads))
        #     starts = list(filter(lambda s: s is not None, starts))
        #     reverses = list(filter(lambda r: r is not None, reverses))
        # else:
        if _loggable:
            logger.log_it("Checking each k-mer for IUPAC substitution or N content.", "DEBUG")
        for i, x in enumerate(kmer_ids): # Here we remove the k-mer ids where N-content is detected, in case they are needed, you can use kmer_ids prior to this point to build functionality.
            if i in sus:
                kmer_ids[i] = None

        if _loggable:
            logger.log_it("Eliminating suspicious 'sus' k-mers, i.e. those with N-content or IUPAC substitutions", "INFO")


            
        kmer_ids = list(filter(lambda k: k is not None, kmer_ids))
        reads = [] # I'm keeping this in, just in case for some reason the variable names are needed in the 
        starts = []
        reverses = []
        if None in kmer_ids:
            # Actually they were just introduced to be filtered out, instead of deleted
            # Because each deletion whould cange the array index
            # So instead we set ghtme to None, and filter out
            raise ValueError("kmerdb.parse.parsefile encountered an invalid kmer_id. Internal error.")



        if _loggable:
            logger.log_it("{0} 'clean' kmers were identified successfully from {1} input sequences".format(len(kmer_ids), num_recs), "DEBUG")
            logger.log_it("Flatmapped {0} kmers for their metadata aggregation".format(len(kmer_ids), len(starts)), "DEBUG")
        # Assert that all list lengths are equal before adding metadata to k-mers
        # else:
        #     raise RuntimeError("Still have no clue what's going on...")
        # On disk k-mer counting
        # Thank you, this was brilliant
        # https://stackoverflow.com/a/9294062/12855110
        # 01-01-2022 This has been removed in favor of in-memory counting
        num_kmers = len(kmer_ids)

        if num_kmers == 0:
            raise ValueError("No k-mers available to add. Please report to the issue tracker")
        else:
            sys.stderr.write("\nAccumulating all k-mers from this set of records...")
            for kmer in kmer_ids:
                counts[kmer] += 1
        # all_kmer_metadata
        # if all_metadata:
        #     for single_kmer_id, read, start, reverse in kmer_metadata:
        #         all_kmer_metadata[single_kmer_id].append((read, start, reverse))

        recs = [r for r in seqprsr] # The next block of exactly 'b' reads


        if _loggable:
            logger.log_it("Read {0} more records from the {1} seqparser object".format(len(recs), "fasta" if fasta else "fastq"), "INFO")
    # Get nullomers
    # only nullomer counts
    unique_kmers = int(np.count_nonzero(counts))


    all_theoretical_kmer_ids = list(range(N))

    
    
    # FIXME
    num_nullomers = N - unique_kmers
    is_nullomer = np.where(counts == 0)


    assert num_nullomers == len(is_nullomer[0]), "kmerdb.parse module find inconsistencies between two ways of counting nullomers. Internal error."
    
    seqprsr.total_kmers = int(np.sum(counts))
    seqprsr.unique_kmers = unique_kmers
    seqprsr.nullomers = num_nullomers

    
    seqprsr.nullomer_array = np.array(all_theoretical_kmer_ids, dtype="uint64")[is_nullomer]


    if _loggable:
        logger.log_it("\n\n\nFinished counting k-mers from '{0}'...\n\n\n".format(filepath), "INFO")

    return counts, seqprsr.header_dict(), seqprsr.nullomer_array, logger.logs





class SeqParser:
    """
    Largely useless module, needs 3 pieces of information passed back in from the outside.
    This performs ugly decompression of fasta and fastq files, patching the __next__ methods, effectively.
    It allows you to read either fasta or fastq data in blocks, obviously useful for the latter.


    :ivar filepath: The .fastq, .fastq.gz, .fasta, .fasta.gz, .fna, .fna.gz, .fa, or .fa.gz file.
    :ivar num: The number of records to read in from a .fastq
    :ivar k: The choice of k to initialize the calculation of kmer/nullomer counts.
    """
    def __init__(self, filepath:str, k:int, logger, num:int=100000):
        """
        The SeqParser class wraps up some functionality  
        """
        
        if type(filepath) is not str:
            raise TypeError("kmerdb.parse.SeqParser expects a str as its first positional argument")
        elif not os.access(filepath, os.R_OK):
            raise TypeError("kmerdb.parse.SeqParser expects an existing path on the operating system as its first argument")
        elif type(num) is not int:
            raise TypeError("kmerdb.parse.SeqParser expects an int as its second positional argument")
        elif type(k) is not int:
            raise TypeError("kmerdb.parse.SeqParser expects an int as its third positional argument")
        elif logger is None:
            raise ValueError("Invalid logger")

        
        self.k = k
        self.num = num
        self.reads = []
        # Header items
        self.filepath = filepath
        self.md5 = None
        self.sha256 = None
        self.total_reads = 0
        self.total_kmers = 0
        self.unique_kmers = 0
        self.nullomers = 0
        self.nullomer_array = []
        self.compressed = False
        self.fastq = False
        exts = os.path.splitext(filepath)

        if exts[-1] == ".gz":
            self.compressed = True
            nogzexts = os.path.splitext(exts[0])
            if nogzexts[-1] == ".fq" or nogzexts[-1] == ".fastq":
                self.fastq = True
            elif nogzexts[-1] == ".fna" or nogzexts[-1] == ".fasta" or exts[-1] == ".fa":
                self.fastq = False
            else:
                raise ValueError("Cannot parse files of extension '{0}'.\n\nRequires fasta (.fna, .fasta, .fa), fastq (.fq, .fastq), or their gzipped equivalents")
        else: # Must be fasta or fastq uncompressed
            if exts[-1] == ".fq" or exts[-1] == ".fastq":
                self.fastq = True
            elif exts[-1] == ".fna" or exts[-1] == ".fasta" or exts[-1] == ".fa":
                self.fastq = False
            else:
                raise ValueError("Cannot parse files of extension '{0}'.\n\nRequires fasta (.fna, .fasta, .fa), fastq (.fq, .fastq), or their gzipped equivalents")

        # This is a really ugly patch to add appropriate fastq and fasta next behavior.
        if self.fastq:
            self.__class__.__iter__ = self._iter_fastq
            self.__class__.__next__ = self._next_fastq
        else:
            self.__class__.__iter__ = self._iter_fasta
            self.__class__.__next__ = self._next_fasta
                
        if self.compressed:
            if self.fastq:
                sys.stderr.write("Opening gzipped fastq file '{0}'...\n".format(filepath))
                self._handle = SeqIO.parse(gzip.open(self.filepath, 'rt'), "fastq")
            else:
                sys.stderr.write("Opening gzipped fasta file '{0}'...\n".format(filepath))
                self._handle = SeqIO.parse(gzip.open(self.filepath, 'rt'), "fasta")
        else:
            if self.fastq:
                sys.stderr.write("Opening uncompressed fastq file '{0}'...\n".format(filepath))
                self._handle = SeqIO.parse(open(self.filepath, 'r'), "fastq")
            else:
                sys.stderr.write("Opening uncompressed fasta file '{0}'...\n".format(filepath))
                self._handle = SeqIO.parse(open(self.filepath, 'r'), "fasta")
        # Get checksums
        self.md5, self.sha256 = self.__checksum()


    def __checksum(self):
        """Generates md5 and sha256 checksums of a file
        :returns: (md5, sha256)
        :rtype: tuple
        """
        if not os.path.exists(self.filepath):
            raise IOError("kmerdb.parse.SeqParser.__checksum could not find '{}' on the filesystem".format(filepath))
        hash_md5 = hashlib.md5()
        hash_sha256 = hashlib.sha256()
        with open(self.filepath, 'rb') as ifile:
            for chunk in iter(lambda: ifile.read(4096), b""):
                hash_md5.update(chunk)
                hash_sha256.update(chunk)
        return (hash_md5.hexdigest(), hash_sha256.hexdigest())

    def header_dict(self):
        """ Create a header dictionary to convert into YAML to go in the header block(s) of the compression header. Has a schema to be validated, defined in config.py

        :returns: dict
        :rtype: dict

        """
        return {
            "filename": self.filepath,
            "md5": self.md5,
            "sha256": self.sha256,
            "total_reads": self.total_reads,
            "total_kmers": self.total_kmers,
            "unique_kmers": self.unique_kmers or 4**self.k - self.nullomers,
            "nullomers": self.nullomers,
        }
            
    def __exit__(self, exc_type, exc_value, traceback):
        self._handle.close()

    def __enter__(self):
        return self
        
        
    def _iter_fastq(self):
        """A custom iterator method to add to the 'reads' array as iterated upon.
        """
        try:
            for i in range(self.num):
                self.reads.append(next(self._handle))
        except StopIteration as e:
            pass
        except ValueError as e:
            raise e
        return self

    def _next_fastq(self):
        """
        A custom mononucleotide counter
        
        """
        if not len(self.reads):
            raise StopIteration
        else:
            self.total_reads += 1
            read = self.reads.pop()
            return read

    def _iter_fasta(self):
        return self

    def _next_fasta(self):

        seq = next(self._handle)
        self.total_reads += 1
        sys.stderr.write("Read {0} sequences from '{1}'...\n".format(self.total_reads, self.filepath))
        return seq
        







class Parseable:
    def __init__(self, arguments, logger=None):
        self.arguments = arguments
        self.logger = logger
            

        
    def parsefile(self, filename):
        """Wrapper function for parse.parsefile to keep arguments succinct for deployment through multiprocessing.Pool
            
        :param filename: the filepath of the fasta(.gz)/fastq(.gz) to process with parsefile -> parse.SeqParser
        :type filename: str
        :returns: (db, m, n)
        """
        return parsefile(filename, self.arguments.k, logger=self.logger)





# def rarefy(arguments):
#     logging.getLogger('matplotlib.font_manager').disabled = True
#     logging.getLogger('matplotlib').setLevel(logging.WARNING)
#     from io import TextIOWrapper
#     import numpy as np
#     import pandas as pd
#     import matplotlib.pyplot as plt

    
#     from ecopy.diversity import rarefy
#     import ecopy as ep
#     from kdb import config


#     if isinstance(arguments.input, argparse.FileType) or isinstance(arguments.input, TextIOWrapper):
#         df = pd.read_csv(arguments.input, sep=arguments.delimiter)
#         column_names = df.columns
#         df = df.transpose()
#     elif arguments.input is None:
#         df = pd.read_csv(sys.stdin, sep=arguments.delimiter)
#         column_names = df.columns
#         df = df.transpose()
#     else:
#         logger.error("An unknown IO type was detected in bin/kdb.cluster()")
#         sys.exit(1)

#     sys.stderr.write(config.DEFAULT_MASTHEAD)
#     if logger.level == logging.DEBUG:
#         sys.stderr.write(config.DEBUG_MASTHEAD)
#     sys.stderr.write(config.RAREFY_MASTHEAD)


#     num_samples, num_features = df.shape
#     t0 = time.time()

#     logger.warning("Multiple files: Untested behavior with ecopy.diversity.rarefy")
#     test = rarefy(df, config.ecopy_rarefaction_fig_filepath, 'rarecurve')
#     #arguments.output.write(test)
#     if arguments.output is None:
#         print(test)
#     else:
#         arguments.output.write(test)

#     sys.stderr.write(config.DONE)



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''


import sys
import os
import copy
import gzip
import yaml
import numpy as np
#import jsonschema
sys.path.append('..')

from kmerdb import fileutil


import logging
logger = logging.getLogger(__file__)



def open(filepath:str=None, indexfilepath:str=None, idx:list=None, mode:str="u", force:bool=False, DEBUG:bool=False):
    if filepath is not None and type(filepath) is not str:
        raise TypeError("kmerdb.index.open expects a str for keyword argument 'filepath'")
    elif indexfilepath is not None and type(indexfilepath) is not str:
        raise TypeError("kmerdb.index.open expects a str for argument 'indexfilepath'")
    elif mode is None or type(mode) is not str:
        raise TypeError("kmerdb.index.open expects a str for argument 'mode'")
    elif force is not None and type(force) is not bool:
        raise TypeError("kmerdb.index.open expects a bool for argument 'force'")
    elif DEBUG is not None and type(DEBUG) is not bool:
        raise TypeError("kmerdb.index.open expects a bool for argument 'DEBUG'")

    if idx is None and ("x" not in mode and "w" not in mode):
        raise ValueError("Index data cannot be None if in create/write/overwrite mode")
    elif idx is not None and type(idx) is not list:
        raise TypeError("kmerdb.index.open expects a list for argument 'idx'")
    elif idx is not None and not all(type(i) is tuple for i in idx):
        raise TypeError("kmerdb.index.open expects a list of tuples for argument 'idx'")
        
    if "w" in mode and idx is not None and type(idx) is not np.ndarray:
        raise TypeError("kmerdb.index.open in create mode expects an int as its first positional argument")
    elif ("w" in mode or "x" in mode) and (kdbfile is None or not isinstance(kdbfile, fileutil.KDBReader)):
        raise ValueError("kmerdb.index.open could not create KDBReader")
            
    modes = set(list(mode.lower()))
    if modes - set("xrwbtu") or len(mode) > len(modes):
        raise ValueError("invalid mode: {}".format(mode))

    unknown = "u" in modes
    creating = "x" in modes
    reading  = "r" in modes
    writing  = "w" in modes
    binary   = "b" in modes
    text     = "t" in modes




    if "u" in mode.lower():
        if filepath is None or type(filepath) is not str:
            raise TypeError("kmerdb.index.open expects a str as its first positional argument")
        elif os.access(filepath, os.R_OK):
            logger.debug("Existing .kdb file found at '{0}'".format(filepath))
            
            if os.access(indexfilepath, os.R_OK):
                logger.debug("Additionally, index file (.kdbi) found at '{0}'".format(indexfilepath))
                if force is True:
                    logger.warning("--force flag provided. Forcing overwrite of index file!")
                    mode = "xt"
                else:
                    mode = "r"
                    raise ValueError("kmerdb.index.open: refusing to write over existing index data")

            else:
                logger.debug("No index data found. Assuming create mode")
                mode = "xt"
        elif not os.access(filepath, os.R_OK):
            raise ValueError("No .kdb data found for indexing.")
            
        else:
            raise ValueError("Could not determine mode from file data and invocation. Internal Error.")
    elif "w" in mode.lower():
        logger.error("Write mode is deprecated. Running in 'create' mode")
        logger.info("Externally provided index. Using write mode (deprecated)")
        raise RuntimeError("Internal error. Write mode is not supported")

        
    if text and binary:
        raise ValueError("can't have text and binary mode simultaneously")
    elif (creating or writing) and reading:
        raise ValueError("must have exactly one or read/write")

        
    if "r" in mode.lower():
        return IndexReader(filepath) # Will attempt to read this as an index file
    elif "x" in mode.lower() or set(list("xw")).intersection(modes):
        logger.info("Building index from file and constructing (mode='x') new .kdbi file")
        write_index(idx, indexfilepath, DEBUG=DEBUG) 



class IndexReader:
    def __init__(self, indexfile:str):
        if type(indexfile) is not str:
            raise TypeError("kmerdb.index.IndexReader expects a str as its first positional argument")
        elif not os.path.exists(indexfile):
            raise TypeError("kmerdb.index.IndexReader expects an existing .kdbi index file as its first positional argument")
        elif not is_gz_file(indexfile):
            raise IOError("kmerdb.index.IndexReader expects a gzip compressed index file as its first positional argument")
        
        idx = []

        #i = 0
        with gzip.open(indexfile, 'rt') as ifile:

            i = 0
            for line in ifile:
                j, kmer_id, line_offset = [int(i) for i in line.rstrip().split("\t")]
                idx.append(line_offset)
                i += 1
        self.index = np.array(idx, dtype="uint64")

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        del self.index
        return


def write_index(index:list, indexfile:str, DEBUG=False):
    if index is None or not(type(t) is tuple for t in index) or not(all(type(i) is int for i in t) for t in index):
        raise TypeError("kmerdb.index.write_index expects .kdb index data as its first positional argument")
    elif type(indexfile) is not str:
        raise TypeError("kmerdb.index.write_index expects a str as its second positional argument")
    elif os.path.exists(indexfile):
        logger.warning("kmerdb.index.write_index is overwriting an existing index '{0}'...".format(indexfile))
        mode = "wt"
    else:
        mode = "xt"

    # print(index)
    # raise RuntimeError("Printed index data")
        
    with gzip.open(indexfile, mode) as ofile:
        for i, offset_tuple in enumerate(index):
            kmer_id, block_offset = offset_tuple
            # File format is index, kmer_id, block_offset
            ofile.write("{0}\t{1}\t{2}\n".format(i, kmer_id, block_offset))




'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''
import sys
import os

import yaml
from collections import OrderedDict

import jsonschema


from kmerdb import config, util



yaml.add_representer(OrderedDict, util.represent_yaml_from_collections_dot_OrderedDict)

default_logline_choices = (20, 50, 100, 200)
PINNED_ISSUES = (132, 133, 137)

PROGRAM_BANNER = """
**** 
 o-O      |||
o---O     |||             [|[          kmerdb           ]|]
O---o     |||
 O-o      |||        version :     v{0}
  O       |||
 o-O      |||        GitHub  : https://github.com/MatthewRalston/kmerdb/issues
o---O     |||         PyPI   : https://pypi.org/project/kmerdb/
O---o     |||      Website   : https://matthewralston.github.io/kmerdb
 O-o      |||
""".format(config.VERSION)
INTERPRETER = "                                                                       lang :         python\n"
# hardcoded

GITHUB_LOGO = """
 .--------------------------------------------------.
 |                 .mmMMMMMMMMMMMMMmm.              |
 |             .mMMMMMMMMMMMMMMMMMMMMMMMm.          |
 |          .mMMMMMMMMMMMMMMMMMMMMMMMMMMMMMm.       |
 |        .MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM.     |
 |      .MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM.   |
 |     MMMMMMMM'  `"MMMMM"""""""MMMM""`  'MMMMMMMM  |
 |    MMMMMMMMM                           MMMMMMMMM |
 |   MMMMMMMMMM:                         :MMMMMMMMMM|
 |  .MMMMMMMMMM                           MMMMMMMMMM.
 |  MMMMMMMMM"                             "MMMMMMMMM
 |  MMMMMMMMM                               MMMMMMMMM
 |  MMMMMMMMM                               MMMMMMMMM
 |  MMMMMMMMMM                             MMMMMMMMMM
 |  `MMMMMMMMMM                           MMMMMMMMMMM
 |   MMMMMMMMMMMM.                     .MMMMMMMMMMMMM
 |    MMMMMM  MMMMMMMMMM         MMMMMMMMMMMMMMMMMMM|
 |     MMMMMM  'MMMMMMM           MMMMMMMMMMMMMMMM` |
 |      `MMMMMM  "MMMMM           MMMMMMMMMMMMMM`   |
 |        `MMMMMm                 MMMMMMMMMMMM`     |
 |          `"MMMMMMMMM           MMMMMMMMM"`       |
 |             `"MMMMMM           MMMMMM"`          |
 |                 `""M           M""`              |
 '--------------------------------------------------'


"""

THREE_LINES = """



"""


DNA_SPACER_1 = """
=================================
=================================


O       o O       o O       o O
| O   o | | O   o | | O   o | | O
| | O | | | | O | | | | O | | | |
| o   O | | o   O | | o   O | | o
o       O o       O o       O O


=================================
"""

DNA_SPACER_lol = """
=================================
=================================
Carbon rules everything around me

O       o O       o O       o O
| O   o | | O   o | | O   o | | O
| | O | | | | O | | | | O | | | |
| o   O | | o   O | | o   O | | o
o       O o       O o       O O


=================================
"""



DNA_COLUMN_1 = """
O---o
 O-o
  O
 o-O
o---O
O---o
 O-o
  O
 o-O
o---O
O---o
 O-o
  O
 o-O
o---O
"""

GITHUB_PROJECT_BANNER = """
==============================================================
                  ||      G i t H u b     ||
==============================================================
                         Repo: kmerdb
               Feature branch: interface

Issue Tracker: https://github.com/MatthewRalston/kmerdb/issues
-------------------------------------------------------
"""



PINNED_ISSUE = """
                 Pinned issue: {0}
""".format(", ".join(list(map(str, PINNED_ISSUES))))





"""
===============================
COMMAND INFO
===============================
"""





command_1_name, command_2_name, command_3_name, command_4_name, command_5_name, command_6_name, command_7_name, command_8_name, command_9_name, command_10_name, command_11_name, command_12_name, command_13_name = config.subcommands

COMMANDS = [
    command_1_name, #"profile",
    command_2_name, #"make_graph",
    command_3_name, #"get_matrix",
    command_4_name, #"distance",
    command_5_name, #"view",
    command_6_name, #"header",
    command_7_name, #"kmeans",
    command_8_name, #"hierarchical",
    command_9_name, #"index",
    command_10_name, #"shuf",
    command_11_name, #"usage",
    command_12_name, #"help"
    command_13_name #"version"
]




        
command_1_description = "create a k-mer count vector from fasta input. g-zipped/.bgzf   -  4 column table with YAML header."
command_1_description_long = """

      :           4 columns output table       :    [   row idx   |   k-mer id  |    count   |  frequency  ]

      :    print a k-mer count vector to STDOUT





           Finally, print a summary of k-mers processed.



>kmerdb profile -k 12 test/data/Cdifficile_R3.fa.gz output.12.kdb


...


-------------------------+
                         |
      Example:           |
                         |
-------------------------+

Final stats:


Total k-mers processed: 4093136
Unique nullomer count:   14509305
Unique 12-mer count:     2267911
Theoretical 12-mer number (4^12):     16777216
==============================



==============================
------------------------------

Output file:

==============================

version: 0.8.0
metadata_blocks: 1
k: 12
total_kmers: 4093136
unique_kmers: 2267911
unique_nullomers: 14509305
metadata: false
sorted: false
kmer_ids_dtype: uint64
profile_dtype: uint64
count_dtype: uint64
frequencies_dtype: float64
tags: []
files:
- filename: test/data/Cdifficile_R3.fasta.gz
  md5: ed09041ad0e991a8037edd0f05643571
  sha256: f2978ff92a19b82d569b10cfe3602eb37a613d6e48db22af82126799beefb1e6
  total_reads: 1
  total_kmers: 4093136
  unique_kmers: 2267911
  nullomers: 14509305

========================



                  +=============+====================+====================+=================================+
                  <    row idx  |  k-mer id          |  k-mer count       |  k-mer frequency        >                        
                  |             |                    |                    |                                 |
                  |             +
                  |
                  |
                  |
                  |
                  |




"""


command_1_parameters = "use < -k > for k-mer size, --quiet to stop STDOUT redundancy, -vv for debug level logging. -v for info."
command_1_inputs = "Input file can be < .fasta | .fastq | .fa.gz >. Last positional arg is output, a gzipped .tsv of k-mer counts and frequencies"
command_1_usage = "kmerdb profile -k $K --quiet <example_1.fa.gz> [example_2.fq.gz] <output_count_vector_file.12.kdb>"
COMMAND_1_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb profile -k 12 input_1.fa output.12.kdb

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]



""".format(command_1_name, command_1_description, command_1_description_long, command_1_inputs, command_1_parameters, command_1_usage)

COMMAND_1_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
        {
            "name"  : "k",
            "type"  : "parameter",
            "value" : "choice of k-mer size parameter 'k'.",
        },
        {
            "name" : "sorted",
            "type" : "flag",
            "value": "descending in k-mer count"
        },
        {
            "name"  : "quiet",
            "type"  : "flag",
            "value" : "write additional debug level information to stderr?"
        }
    ]
})

COMMAND_1_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name"  : "<.fasta|.fastq>",
            "type"  : "array",
            "value" : "gzipped or uncompressed input .fasta or .fastq file(s)"
        },
        {
            "name"  : "<output>.$K.kdb - a k-mer count vector in a 4 column output table. zlib compatible",
            "type"  : "file",
            "value" : "Output file. K-mer count vector in a 4 column table (index number, k-mer id, count, and frequency)."
        }
    ]
})


COMMAND_1_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "k-mer count array produced as sequences are read by sliding window approach. (Un)compressed support for .fa/.fq.",
            "shortname": "parallel OOP sliding window k-mer shredding",
            "description": "N = 4^k count-vector from one or more sequence files. (index, k-mer id, count, and frequency [as float64])"
        }),
        OrderedDict({
            "name": "k-mers are tallied, and metadata merged from the input files",
            "shortname": "merge counts, metadata from across inputs",
            "description": "a final metadata structure and output metrics are collected for display to the user."
        })

            ]
})


COMMAND_1_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read input file(s) from filesystem into k-mer arrays",
            "shortname": "shred inputs into k-mer count arrays",
            "description": "shred input sequences into k-mer count vector",
        }),
        OrderedDict({
            "name": "collate the count array",
            "shortname": "collate the count array",
            "description": "aggregate counts from different input files"
        }),
        OrderedDict({
            "name": "merge k-mer arrays and aggregate metadata",
            "shortname": "merge k-mer count arrays for aggregate metadata (header)",
            "description": "merge counts of nullomers, unique kmers, and total kmers."
        }),
        OrderedDict({
            "name": "print 'table' Final stats and close output file",
            "shortname": "metrics and shutdown",
            "description": "print final statistics, typically metadata values, and ensure file is closed."
        })

    ]

})


command_2_description = "create a edge list in (block) .gz format from .fasta|.fa or .fastq format."
command_2_description_long = """


   :     4 column output : [ row idx | k-mer id node #1 | k-mer id node #2 | edge weight (adjacency count) ]

   :  make a deBruijn graph, count the number of k-mer adjacencies,  printing the edge list to STDOUT




                  +=============+====================+====================+=================================+
                  <    row idx  |  k-mer id node #1  |  k-mer id node #2  |  edge weight (adjacency count)  >
                  |             |                    |                    |                                 |
                  |             +
                  |
                  |
                  |
                  |
                  |
"""
        
command_2_parameters = "uses < -k > for k-mer size, --quiet to reduce runtime, -v, -vv to control logging. --"
command_2_inputs = "Input file can .fastq (or .fa).   - gzip.  Output is a weighted edge list in .kdb format (gzipped .csv with YAML header)"
command_2_usage = "kmerdb graph -k $K --quiet <input_1.fa.gz> [input_2.fq.gz] <output_edge_list_file.12.kdbg>" 


COMMAND_2_BANNER = """



                          [ name ] :         {0}

                   description : {1}

{2}




--------------------------


                    kmerdb {0} -k 12 input_1.fa [example_2.fastq] output.12.kdbg

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}











""".format(command_2_name, command_2_description, command_2_description_long, command_2_inputs, command_2_parameters, command_2_usage)

        

        
COMMAND_2_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
        {
            "name"  : "k",
            "type"  : "int",
            "value" : "choice of k-mer size"
        },
        {
            "name"  : "quiet",
            "type"  : "flag",
            "value" : "Write additional debug level information to stderr?"
        }
    ]
})
        
        

COMMAND_2_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name"  : "<.fasta|.fastq>",
            "type"  : "array",
            "value" : "gzipped or uncompressed input .fasta or .fastq file(s)"
        },
        {
            "name"  : ".kdbg",
            "type"  : "file",
            "value" : "Output edge-list filepath."
        }
    ]
})


COMMAND_2_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "k-mer count arrays, linear, produced as file is read through sliding window. (Un)compressed support for .fa/.fq.",
            "shortname": "parallel faux-OP sliding window k-mer shredding",
            "description": "Sequential k-mers from the input .fq|.fa files are added to the De Bruijn graph. In the case of secondary+ sequences in the .fa or considering NGS (.fq) data, non-adjacent k-mers are pruned with a warning. Summary statistics for the entire file are given for each file read, + a transparent data structure."
        }),
        OrderedDict({
            "name": "k-mer neighbors assessed and tallied, creates a unsorted edge list, with weights",
            "shortname": "weighted undirected graph",
            "description": "an edge list of a De Bruijn graph is generated from all k-mers in the forward direction of .fa/.fq sequences/reads. i.e. only truly neighboring k-mers in the sequence data are added to the tally of the k-mer nodes of the de Bruijn graph and the edges provided by the data."
        })
    ]
})


COMMAND_2_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read input file(s) from filesystem into k-mer arrays",
            "shortname": "shred inputs into k-mer count arrays",
            "description": "shred input sequences into k-mer count vector",
        }),
        OrderedDict({
            "name": "merge k-mer arrays and aggregate metadata",
            "shortname": "merge k-mer count arrays for aggregate metadata (header)",
            "description": "merge counts of nullomers, unique kmers, and total kmers."
        }),
        OrderedDict({
            "name": "collate the weighted edge lists after reading multiple files. Output data consists of a edge_list, analogous metadata-header as YAML, kmer_counts, and nullomer_ids.",
            "shortname": "extract undirected weighted graph",
            "description": "consists of info from a .kdb file and a .kdbg file. The node IDs, the edges, and the number of times the pair was observed from forward sequences in the provided dataset"
        }),
        OrderedDict({
            "name": "print 'table' Final stats and close output file",
            "shortname": "metrics and shutdown",
            "description": "print final statistics, typically metadata values, and ensure file is closed."
        })
        
            ]
})
        

command_3_description = "Aggregate k-mer count vectors into k-mer count matrix. Output is plain text .tsv"
command_3_description_long = """








...





"""
command_3_parameters = "Use PCA or tSNE for dim. reduction. Use DESeq2 for count vector sample normalization. Use 'pass' to just aggregate count vectors into .tsv"
command_3_inputs = "Several input files with the same [-k]. Or a k-mer count matrix as .tsv. Use 'STDIN' in place of an input file to read .tsv from STDIN in a Unix pipe. Prints .tsv to stdout."
command_3_usage = "kmerdb matrix <pass|PCA|tSNE|DESeq2> [input1.12.kdb input2.12.kdb ...] [OR] kmerdb matrix DESeq2 <input_count_matrix.tsv> [OR] kmerdb matrix PCA -k 5 STDIN"

        
COMMAND_3_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb matrix <pass|PCA|tSNE|DESeq2> input1.12.kdb input2.12.kdb ...

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]
""".format(command_3_name, command_3_description, command_3_description_long, command_3_inputs, command_3_parameters, command_3_usage)
        
        
COMMAND_3_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
        {
            "name" : "function",
            "type": "parameter",
            "value" : "dimensionality reduction, DESeq2 normalization (via rpy2), or pass - no-transformation, or (deprecated) k-mer Frequency matrix. [PCA, tSNE, DESeq2, pass, Frequency]",
        },
        {
            "name" : "n",
            "type":  "int",
            "value": "dimension choise for scipy PCA or tSNE"
        },
        {
            "name": "perplexity",
            "type": "int",
            "value" : "perplexity parameter for scipy tSNE"
        },
        {
            "name": "no-normalized-ints",
            "type": "flag",
            "value": "do not round counts to nearest integer when importing from RPy2 from DESeq2 normalization"
        }
    ]
})


COMMAND_3_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name": "kdbfiles",
            "type": "file(s)",
            "value": ".kdb count vector files produced from command #1 in the profile <-> k-mer count matrix <-> count distance matrix pipeline"
        },
        {
            "name": "input.tsv",
            "type": "file",
            "value": "A table of csv/tsv file counts, possibly for dimensionality reduction."
        },
        {
            "name": "STDIN",
            "type": "flag",
            "value": "Read from STDIN, possibly in a unix pipe."
        }
        
            ]
})

COMMAND_3_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read/import datasets",
            "shortname": "read/import multiple count vectors",
            "description" : "count vectors in columns of csv/tsv input over 'STDIN', .tsv/csv file (with --delimiter option), or multiple .kdb files."
        }),
        OrderedDict({
            "name": "Pandas library .csv sanitation",
            "shortname": "validate input",
            "description": "sanitize and structure validation on inputs (TODO: 4/5/24 not done yet. Uses a pandas csv loader on tsv/csv files or 'STDIN' to read a matrix/DataFrame input from a standard input stream. Otherwise collates multiple .kdb count vectors and coerces the input into a NumPy array otherwise failure.)"
        }),
        OrderedDict({
            "name": "transformation on data matrix",
            "shortname": "transformation on data matrix",
            "description": "Apply 'transformation' to input matrix. Not a distance calculation, but a restructuring of the data. Options are PCA (+scipy), t-stochastic neighbor embedding(scipy), 'DESeq2' median-of-ratios normalization, or retrieve frequencies for related applications."
        })
    ]
})

COMMAND_3_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read input file(s) from filesystem into k-mer arrays",
            "shortname": "read .kdb files into k-mer count arrays",
            "description": "read multiple .kdb files into k-mer count matrix",
        }),
        OrderedDict({
            "name": "parse .csv into Pandas DataFrame",
            "shortname": "Pandas .csv import from input file or STDIN",
            "description": "merge counts of nullomers, unique kmers, and total kmers."
        }),
        OrderedDict({
            "name": "Apply PCA dimensionality reduction with provided dimensionality n",
            "shortname": "PCA dim reduction",
            "description": "Uses scipy PCA to apply dimensionality reduction on input matrix"
        }),
        OrderedDict({
            "name": "Apply t-SNE dimensionality reduction with provided dimensionality n",
            "shortname": "t-SNE dim reduction",
            "description": "Uses scipy t-SNE to apply dimensionality reduction on input matrix"
        }),
        OrderedDict({
            "name": "Apply DESeq-2 'median-of-ratios' count normalization",
            "shortname": "DESeq-2 normalize",
            "description": "Uses rpy2 to call the R server and pass the data matrix for DESeq-2 normalization"

                    }),
        OrderedDict({
            "name": "pass unnormalized counts",
            "shortname": "unnormalized data",
            "description": "Passes the unnormalized data in Data Frame tsv format to STDOUT/file"
        })

    ]

})



command_4_description = "Create distance matrix from input count matrix (.tsv). Output is plain text .tsv"
command_4_description_long = """








...





"""
command_4_parameters = "Choose a distance metric. 23 distances available. 'pearson' 'correlation' and 'spearman' are relevant for k-mer count vector similarity."
command_4_inputs = "Several input files with the same [-k]. Or a k-mer count matrix as .tsv. Use 'STDIN' in place of an input file to read .tsv from STDIN in a Unix pipe. Prints .tsv to stdout."
command_4_usage = "kmerdb distance  [input1.12.kdb input2.12.kdb ...] [OR] kmerdb distance correlation <input.tsv> [OR] kmerdb distance spearman STDIN"

        
COMMAND_4_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb distance pearson [input1.12.kdb input2.12.kdb ...]  OR kmerdb distance pearson <input.tsv> or kmerdb distance pearson STDIN

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]
""".format(command_4_name, command_4_description, command_4_description_long, command_4_inputs, command_4_parameters, command_4_usage)

        

COMMAND_4_PARAMS = OrderedDict({
    "name": "arguments",
            "type": "array",
    "items": [
        {
            "name" : "distance method",
            "type": "parameter",
            "value" : "braycurtis,canberra,chebyshev,cityblock,correlation,cosine,dice,euclidean,hamming,jaccard,jensenshannon,kulsinski,mahalanobis,matching,minkowski,pearson,rogerstanimotorusselrao,seuclidean,sokalmichener,sokalsneath,spearman,sqeuclidean,yule distances. 'spearman' and 'correlation' distances provided by scipy. 'pearson' is a custom Cython correlation coefficient: ssxy/sqrt(ssxx^2 x ssyy^2) | all others provided by scipy",
        },
    ]
})

        
COMMAND_4_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name": "kdbfiles",
            "type": "file(s)",
            "value": ".kdb count vector files produced from kmerdb profile"
        },
        {
            "name": "input.tsv",
            "type": "file",
            "value": "A table of csv/tsv file counts, possibly for dimensionality reduction."
        },
        {
            "name": "STDIN",
            "type": "flag",
            "value": "Read from STDIN, possibly in a unix pipe."
        }
        
            ]
})

COMMAND_4_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read/import datasets",
            "shortname": "read/import multiple count vectors",
            "description" : "count vectors in columns of csv/tsv input over 'STDIN', .tsv/csv file (with --delimiter option), or multiple .kdb files."
        }),
        OrderedDict({
            "name": "Pandas library .csv sanitation",
            "shortname": "validate input",
            "description": "sanitize and structure validation on inputs (TODO: 4/5/24 not done yet. Uses a pandas csv loader on tsv or STDIN input. Otherwise collates multiple .kdb count vectors and coerces the input into a NumPy array otherwise failure.)"
        }),
        OrderedDict({
            "name": "generates distance matrix",
            "shortname": "generates distance matrix",
            "description": "Choose 'distance method' to apply to input matrix. Available SciPy distances include 'braycurtis,canberra,chebyshev,cityblock,correlation,cosine,dice,euclidean,hamming,jaccard,jensenshannon,kulsinski,mahalanobis,matching,minkowski,pearson,rogerstanimotorusselrao,seuclidean,sokalmichener,sokalsneath,spearman,sqeuclidean,yule' distances. 'spearman' and 'correlation' distances provided by scipy. 'pearson' is a custom Cython correlation coefficient: ssxy/sqrt(ssxx^2 x ssyy^2) | all others provided by scipy. All distances availalable on -h|--help"
        })
    ]
})
    
COMMAND_4_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read input file(s) from filesystem into k-mer arrays",
            "shortname": "read .kdb files into k-mer count arrays",
            "description": "read multiple .kdb files into k-mer count matrix",
        }),
        OrderedDict({
            "name": "parse .csv into Pandas DataFrame",
            "shortname": "Pandas .csv import from input file or STDIN",
            "description": "merge counts of nullomers, unique kmers, and total kmers."
        }),
        OrderedDict({
            "name": "calculate distance matrix from input data matrix",
            "shortname": "distance matrix generation",
            "description": "Use input data matrix to create a distance matrix, produced by SciPy and a choice of distance method"
        })

    ]
    
})

















command_5_description = "Decompress and view contents of .kdb(g) file "
command_5_description_long = """








...





"""
command_5_parameters = "Use -H|--header to additionally print header."
command_5_inputs = "Input may be a v{0} count vector (.kdb) or edge list (.kdbg) file.".format(config.VERSION)
command_5_usage = "kmerdb view -H <input_1.12.kdb>"

        
COMMAND_5_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb view -H <input_1.12.kdb>

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]
""".format(command_5_name, command_5_description, command_5_description_long, command_5_inputs, command_5_parameters, command_5_usage)

        

COMMAND_5_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
        {
            "name" : "header flag",
            "type": "parameter",
            "value" : "Use -H|--header to print YAML metadata header to STDOUT before table section"
        },
    ]
})

        
COMMAND_5_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name": "k-mer database file (.kdb)",
            "type": "file",
            "value": ".kdb count vector files produced from kmerdb profile"
        }
        
            ]
})

COMMAND_5_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read inputs: parse and validate input file's header and table",
            "shortname": "read input data into count vectors",
            "description" : "parse input data, a custom format, and load data into memory"
        }),
        OrderedDict({
            "name": "print file contents",
            "shortname": "print to STDOUT",
            "description": ""
        })
    ]
})
    
COMMAND_5_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read input file from filesystem into k-mer count vector",
            "shortname": "read .kdb file into k-mer count vector",
            "description": "read a single .kdb file into a k-mer count vector",
        }),
        OrderedDict({
            "name": "print file contents to STDOUT",
            "shortname": "decompress and print input file",
            "description": ""
        })

    ]

})




command_6_description = "Show YAML metadata header from .kdb(g) file"
command_6_description_long = ""
command_6_parameters = "N/A"
command_6_inputs = "Input may be a v{0} count vector (.kdb) or edge list (.kdbg) file.".format(config.VERSION)
command_6_usage = "kmerdb header <input_1.12.kdb>"

        
COMMAND_6_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb header <input_1.12.kdb>

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]
""".format(command_6_name, command_6_description, command_6_description_long, command_6_inputs, command_6_parameters, command_6_usage)

        

COMMAND_6_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
        {
            "name" : "json",
            "type": "parameter",
            "value" : "Print results in JSON format"
        },
    ]
})

        
COMMAND_6_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name": "k-mer database file (.kdb)",
            "type": "file",
            "value": ".kdb count vector file produced from kmerdb profile"
        }
        
            ]
})

COMMAND_6_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read inputs: parse and validate input file's header and table",
            "shortname": "read input data into count vectors",
            "description" : "parse input data, a custom format, and load data into memory"
        }),
        OrderedDict({
            "name": "print YAML metadata header",
            "shortname": "print to STDOUT",
            "description": ""
        })
    ]
})
    
COMMAND_6_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "read input file from filesystem into k-mer count vector",
            "shortname": "read .kdb file into k-mer count vector",
            "description": "read a single .kdb file into a k-mer count vector",
        }),
        OrderedDict({
            "name": "print file contents to STDOUT",
            "shortname": "decompress and print input file",
            "description": ""
        })

    ]

})


















command_7_description = "K-means clustering with biopython or scikit-learn"
command_7_description_long = "Produces eblow graph if k is not determined a-priori. Uses matplotlib for graphics."
command_7_parameters = "Use -k to control clustering with k-means. Choice of 'sklearn' or 'biopython' k-means clustering. Choice of distance metrics offered by kcluster"
command_7_inputs = "Input is a .tsv file. Use STDIN to read input from standard input".format(config.VERSION)
command_7_usage = "kmerdb kmeans -k 5 -i <count_or_distance_matrix.tsv> sklearn"

        
COMMAND_7_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb kmeans -k 5 <count_matrix.tsv> biopython

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]
""".format(command_7_name, command_7_description, command_7_description_long, command_7_inputs, command_7_parameters, command_7_usage)

        

COMMAND_7_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
        {
            "name" : "k",
            "type": "parameter",
            "value" : "Controls k-means clustering"
        },
    ]
})

        
COMMAND_7_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name": "Input data matrix",
            "type": "file",
            "value": "Input may be a count matrix or distance/correlation matrix."
        }
        
            ]
})

COMMAND_7_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "Parse input count matrix or distance/correlation matrix",
            "shortname": "read input data matrix",
            "description" : "load data into memory from input file"
        }),
        OrderedDict({
            "name": "Use biopython or scikit-learn for k-means clutsering",
            "shortname": "kmeans clustering",
            "description": "creates elbow graph if choice of k is ambiguous, and other output graphics from distance matrix"
        })
    ]
})
    
COMMAND_7_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "Parse .csv file",
            "shortname": "Loads data into Pandas DataFrame",
            "description": "Infers datatype from .csv data",
        }),
        OrderedDict({
            "name": "Run k-means clustering",
            "shortname": "k-means clustering",
            "description": "Uses biopython or scikit-learn for k-means clustering, and matplotlib for graphics"
        })

    ]

})





command_8_description = "Hierarchical clustering with biopython"
command_8_description_long = "Uses matplotlib for graphics."
command_8_parameters = "-m|--method determines linkage fitting (see --help for details)"
command_8_inputs = "Input is a .tsv file. Use STDIN to read input from standard input".format(config.VERSION)
command_8_usage = "kmerdb hierarchical -m ward -i <count_or_distance_matrix.tsv>"

        
COMMAND_8_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb kmeans -k 5 <count_matrix.tsv> biopython

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]
""".format(command_8_name, command_8_description, command_8_description_long, command_8_inputs, command_8_parameters, command_8_usage)

        

COMMAND_8_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
        {
            "name" : "method",
            "type": "parameter",
            "value" : "Linkage method for hierarchical clustering. Choices are single,complete,average,weighted,centroid,median,ward"
        },
    ]
})

        
COMMAND_8_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name": "Input data matrix",
            "type": "file",
            "value": "Input may be a count matrix or distance/correlation matrix."
        }
        
    ]
})

COMMAND_8_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "Parse input count matrix or distance/correlation matrix",
            "shortname": "read input data matrix",
            "description" : "load data into memory from input file"
        }),
        OrderedDict({
            "name": "Use biopython for hierarchical clutsering",
            "shortname": "hierarchical clustering",
            "description": "Uses Bio.Cluster.treecluster for hierarchical clustering. Produces a phylip UPGMA tree."
        })
    ]
})
    
COMMAND_8_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "Parse .csv file",
            "shortname": "Loads data into Pandas DataFrame",
            "description": "Infers datatype from .csv data",
        }),
        OrderedDict({
            "name": "Run hierarchical clustering",
            "shortname": "hierarchical clustering",
            "description": "Uses biopython hierarchical clustering. Produces a UPGMA tree in Phylip format"
        })

    ]

})























command_9_description = "Create an index file (deprecated)"
command_9_description_long = ""
command_9_parameters = "N/A"
command_9_inputs = "Input is a v{0} .kdb count vector file".format(config.VERSION)
command_9_usage = "kmerdb index <kmer_count_vector.kdb>"

        
COMMAND_9_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb index <input.kdb>

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]
""".format(command_9_name, command_9_description, command_9_description_long, command_9_inputs, command_9_parameters, command_9_usage)

        

COMMAND_9_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
    ]
})

        
COMMAND_9_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name": "Input k-mer count vector (.kdb) file",
            "type": "file",
            "value": "File to index."
        }
        
            ]
})

COMMAND_9_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "Parse input k-mer count vector .kdb file",
            "shortname": "read input file",
            "description" : "load data into memory from input file"
        }),
        OrderedDict({
            "name": "Create index file",
            "shortname": "index input file",
            "description": "Create a plaintext index file for on-disk operations when memory may be an issue. (Deprecated)"
        })
    ]
})
    
COMMAND_9_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "Parse input k-mer db .kdb file",
            "shortname": "Loads data from input file",
            "description": "Validate and parse input data",
        }),
        OrderedDict({
            "name": "Create index file",
            "shortname": "Create index file",
            "description": "Create a plaintext index file (.kdbi) for on disk operations (Deprecated)"
        })

    ]

})


        

command_10_description = "Shuffle k-mer count vector (Deprecated)"
command_10_description_long = ""
command_10_parameters = "N/A"
command_10_inputs = "Input is a v{0} .kdb count vector file".format(config.VERSION)
command_10_usage = "kmerdb shuf <kmer_count_vector.kdb>"

        
COMMAND_10_BANNER = """












[--------------------------------------------------------------------------------------]








                        [  n a m e    ]         :  -   {0}

                   description : {1}

{2}




--------------------------

                    kmerdb shuf <input.kdb>

                    [-]    inputs : 

                           {3}

                    [-]    parameters : 

                           {4}



                    [-]    [ usage ]  :  {5}


















[--------------------------------------------------------------------------------------]
""".format(command_10_name, command_10_description, command_10_description_long, command_10_inputs, command_10_parameters, command_10_usage)

        

COMMAND_10_PARAMS = OrderedDict({
    "name": "arguments",
    "type": "array",
    "items": [
    ]
})

        
COMMAND_10_INPUTS = OrderedDict({
    "name": "inputs",
    "type": "array",
    "items": [
        {
            "name": "Input k-mer count vector (.kdb) file",
            "type": "file",
            "value": "File to shuffle."
        }
        
    ]
})

COMMAND_10_FEATURES = OrderedDict({
    "name": "features",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "Parse input k-mer count vector .kdb file",
            "shortname": "read input file",
            "description" : "load data into memory from input file"
        }),
        OrderedDict({
            "name": "Shuffle k-mer count vector",
            "shortname": "Shuffle k-mer counts",
            "description": "(Deprecated)"
        })
    ]
})
    
COMMAND_10_STEPS = OrderedDict({
    "name": "steps",
    "type": "array",
    "items": [
        OrderedDict({
            "name": "Parse input k-mer db .kdb file",
            "shortname": "Loads data from input file",
            "description": "Validate and parse input data",
        }),
        OrderedDict({
            "name": "Shuffle k-mer count vector",
            "shortname": "Shuffle k-mer counts",
            "description": "(Deprecated)"
        })

    ]

})

ALL_PARAMS = {
    "profile": COMMAND_1_PARAMS["items"],
    "graph": COMMAND_2_PARAMS["items"],
    "matrix": COMMAND_3_PARAMS["items"],
    "distance": COMMAND_4_PARAMS["items"],
    "kmeans": COMMAND_7_PARAMS["items"],
    "hierarchical": COMMAND_8_PARAMS["items"],    
    "view": COMMAND_5_PARAMS["items"],
    "header": COMMAND_6_PARAMS["items"],
    "index": COMMAND_9_PARAMS["items"],
    "shuf": COMMAND_10_PARAMS["items"]
}

ALL_INPUTS = {
    "profile": COMMAND_1_INPUTS["items"],
    "graph": COMMAND_2_INPUTS["items"],
    "matrix": COMMAND_3_INPUTS["items"],
    "distance": COMMAND_4_INPUTS["items"],
    "kmeans": COMMAND_7_INPUTS["items"],
    "hierarchical": COMMAND_8_INPUTS["items"],
    "view": COMMAND_5_INPUTS["items"],
    "header": COMMAND_6_INPUTS["items"],
    "index": COMMAND_9_INPUTS["items"],
    "shuf": COMMAND_10_INPUTS["items"]
}

ALL_FEATURES = {
    "profile": COMMAND_1_FEATURES["items"],
    "graph": COMMAND_2_FEATURES["items"],
    "matrix": COMMAND_3_FEATURES["items"],
    "distance": COMMAND_4_FEATURES["items"],
    "kmeans": COMMAND_7_FEATURES["items"],
    "hierarchical": COMMAND_8_FEATURES["items"],
    "view": COMMAND_5_FEATURES["items"],
    "header": COMMAND_6_FEATURES["items"],
    "index": COMMAND_9_FEATURES["items"],
    "shuf": COMMAND_10_FEATURES["items"]


}


ALL_STEPS = {

    "profile": COMMAND_1_STEPS["items"],
    "graph": COMMAND_2_STEPS["items"],
    "matrix": COMMAND_3_STEPS["items"],
    "distance": COMMAND_4_STEPS["items"],
    "kmeans": COMMAND_7_STEPS["items"],
    "hierarchical": COMMAND_8_STEPS["items"],
    "view": COMMAND_5_STEPS["items"],
    "header": COMMAND_6_STEPS["items"],
    "index": COMMAND_9_STEPS["items"],
    "shuf": COMMAND_10_STEPS["items"]
}




        
class kmerdb_appmap:



    
    def __init__(self, argv, logger=None):


        if logger is not None:
            self.logger = logger
            self.logfile = logger.logfile
        else:
            self.logger = None
            self.logfile = None
        self._loggable = logger is not None
        
        
        self.MODULE_ROOT = os.path.join("..", os.path.dirname(__file__))
        self.COMMAND_FILE = os.path.join(self.MODULE_ROOT, "__init__.py")
        self.PACKAGE_MANAGER = """
                      package manger : pip
                        version      : >= 24.0
        package root : {0}
        exe file     : {1}

                      required packages : {2}
                   development packages : {3}

           ARGV : {4}
        """.format(self.MODULE_ROOT, self.COMMAND_FILE, config.requirements_count, config.requirements_dev_count, argv)
        self.REQUIRES_PYTHON = config.REQUIRES_PYTHON
        self.VERSION_HARDCODED = "                                                                          v :      >= v{0}\n".format(config.REQUIRES_PYTHON)


        
        #
        # loaded_modules
        #

        #
        # dependencies
        #
        #      required:
        #      optional:

        # usage_notes.txt





        


        
        # KMEANS_BANNER = """
        #                   name : kmeans
        #            description : 
        # """

        # HIERARCHICAL_BANNER = """

        # """


        # config.py features
        #SUPPORTED_FEATURES = {}

        # ASCII + RICH.py checklist
        
    def print_verbosity_header(self):
        """
        This prints the verbosity warnings and default logging behavior for the user.
        """

        sys.stderr.write("="*40 + "\n")
        sys.stderr.write("[ DEBUG ] : Default is warning-only logging. [-v,-vv, --debug, --quiet]\n")
        sys.stderr.write("[ INFO  ] : Default is warning-only logging. [-v,-vv, --debug, --quiet]\n")
        sys.stderr.write("="*40 + "\n")
        sys.stderr.write("WARNING. Some features are experimental. Note: you are tracking kmerdb/main. Visit the README.md header, usage section, or quickstart page for additl.\n")
        sys.stderr.write("-"*40 + "\n")

        #print("x..xx")
        #print("xd up ya brakes")

        #
        #
        #...


        #' ...#x'"' ''"'?"    # ... #xdupyabrakexsz '#xdupyabreakesz' '"'"''"""""""...'"'. #XduPyabrakx. '"' ....

        return




    def print_program_header(self):
        """
        Only used in 'usage' method. Spacer at bottom is permissible because this may be used externally to usage.
        """
        
        sys.stderr.write(PROGRAM_BANNER)
        sys.stderr.write(INTERPRETER)
        sys.stderr.write(self.VERSION_HARDCODED)
        sys.stderr.write(self.PACKAGE_MANAGER)

        # Spacer
        sys.stderr.write(DNA_COLUMN_1)

        sys.stderr.write(THREE_LINES)

    def print_graph_header(self):
        """
        'kmerdb usage graph'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.

        """
        
        sys.stderr.write(COMMAND_2_BANNER)

        sys.stderr.write(THREE_LINES)
        
        sys.stderr.write(yaml.dump(COMMAND_2_PARAMS))
        
        sys.stderr.write(THREE_LINES)
        
        sys.stderr.write(yaml.dump(COMMAND_2_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_2_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_2_STEPS))



    def print_profile_header(self):
        """
        'kmerdb usage profile'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.

        """
        sys.stderr.write(COMMAND_1_BANNER)

        sys.stderr.write(THREE_LINES)
            
        sys.stderr.write(yaml.dump(COMMAND_1_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_1_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_1_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_1_STEPS))



    def print_matrix_header(self):
        """
        'kmerdb usage matrix'


        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.
        """

        sys.stderr.write(COMMAND_3_BANNER)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_3_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_3_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_3_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_3_STEPS))


    def print_distance_header(self):
        """

        'kmerdb usage distance'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.
        """

        sys.stderr.write(COMMAND_4_BANNER)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_4_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_4_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_4_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_4_STEPS))





    def print_view_header(self):
        """

        

        'kmerdb usage header'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.


        """

        sys.stderr.write(COMMAND_5_BANNER)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_5_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_5_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_5_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_5_STEPS))


    def print_header_header(self):
        """

        

        'kmerdb usage header'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.


        """

        sys.stderr.write(COMMAND_6_BANNER)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_6_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_6_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_6_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_6_STEPS))



    def print_kmeans_header(self):
        """

        

        'kmerdb usage kmeans'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.


        """

        sys.stderr.write(COMMAND_7_BANNER)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_7_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_7_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_7_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_7_STEPS))


    def print_hierarchical_header(self):
        """

        

        'kmerdb usage hierarchical'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.


        """

        sys.stderr.write(COMMAND_8_BANNER)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_8_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_8_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_8_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_8_STEPS))



        
    def print_index_header(self):
        """

        

        'kmerdb usage index'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.


        """

        sys.stderr.write(COMMAND_9_BANNER)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_9_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_9_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_9_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_9_STEPS))




        
    def print_shuf_header(self):
        """

        

        'kmerdb usage shuf'

        Writes input files, parameters, steps, and features as verbose --help on the parameters effects on the runtime.


        """

        sys.stderr.write(COMMAND_10_BANNER)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_10_PARAMS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_10_INPUTS))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_10_FEATURES))

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(yaml.dump(COMMAND_10_STEPS))



        
        
    def print_github_block(self):
        """
        Github repo block. May be useful to some users when browsing logilfe.
        """


        sys.stderr.write(THREE_LINES)
        
        sys.stderr.write(DNA_SPACER_1)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(GITHUB_LOGO)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(GITHUB_PROJECT_BANNER)

        sys.stderr.write(PINNED_ISSUE)

        sys.stderr.write(THREE_LINES)

        sys.stderr.write(DNA_SPACER_1)

        sys.stderr.write(THREE_LINES)

        
        

    def exit_gracefully(self, e:Exception, subcommand:str=None, step:int=None, feature:int=None, logs:list=None, n_logs:int=None):
        """
        We need to handle exit gracefully. The 'step' and 'feature' categories/flags/ints are passed from __init__ or down its callstack to 
        """
        import traceback


        # if e is None:
        #     raise ValueError("Need an error to exit")
        # elif not isinstance(e, Exception):
        #     raise ValueError("Need an error to exit")

        
        # if n_logs is None or type(n_logs) is not int:
        #     raise TypeError("kmerdb.appmap.exit_gracefully expects the keyword argument n_logs to be a int")
        # elif logs is None or type(logs) is not list:
        #     raise TypeError("kmerdb.appmap.exit_gracefully expects the keyword argument logs to be a list")
        # elif feature is not None and type(feature) is not int:
        #     raise TypeError("kmerdb.appmap.exit_gracefully expects the keyword argument feature to be a int")
        # elif step is not None and type(step) is not int:
        #     raise TypeError("kmerdb.appmap.exit_gracefully expects the keyword argument step to be a int")
        # elif subcommand is not None and type(subcommand) is not str:
        #     raise TypeError("kmerdb.appmap.exit_gracefully expects the keyword argument subcommand to be a str")



        
        N = min(len(logs), n_logs)

        tb = traceback.extract_tb(e.__traceback__)
        last_traceback_FrameSummary = tb[-1]
        error_file_name = last_traceback_FrameSummary.filename
        error_line_number = last_traceback_FrameSummary.lineno

        
        #assert subcommand in config.subcommand_functions, "Unknown subcommand"

        if self._loggable:
            self.logger.log_it("Program error! Collecting error metadata and formatting...", "ERROR")
        # This is the "Error blocks" metadata


        sys.stderr.write("Aggregating program metadata, if this fails without error without the --debug flag, please report to the GitHub issue tracker with the title 'Error summary convenience function'.")
        sys.stderr(subcommand)
        sys.stderr(config.VERSION)
        sys.stderr(config.REQUIRES_PYTHON)
        sys.stderr(feature)
        sys.stderr(ALL_FEATURES[subcommand][feature]["name"])
        sys.stderr(ALL_FEATURES[subcommand][feature]["shortname"])
        sys.stderr(ALL_FEATURES[subcommand][feature]["description"])
        sys.stderr(step)
        sys.stderr(ALL_STEPS[subcommand][step]["name"])
        sys.stderr(ALL_STEPS[subcommand][step]["shortname"])
        sys.stderr(ALL_STEPS[subcommand][step]["description"])
            # The *total* number of logged lines produced by the program and returned to the global 'logs' var in __init__.py
        sys.stderr(self.logfile)
        sys.stderr(str(tb))
        sys.stderr(error_file_name)
        sys.stderr(error_line_number)
        sys.stderr(e.__str__())

        
        e_sum = {
            "subcommand": subcommand,
            "kmerdb-version": config.VERSION,
            "python-version": config.REQUIRES_PYTHON,
            "feature": feature,
            "feature_name": ALL_FEATURES[subcommand][feature]["name"],
            "feature_shortname": ALL_FEATURES[subcommand][feature]["shortname"],
            "feature_description": ALL_FEATURES[subcommand][feature]["description"],
            "step" : step,
            "step_name": ALL_STEPS[subcommand][step]["name"],
            "step_shortname": ALL_STEPS[subcommand][step]["shortname"],
            "step_description": ALL_STEPS[subcommand][step]["description"],
            # The *total* number of logged lines produced by the program and returned to the global 'logs' var in __init__.py
            "log_file": self.logfile,
            "traceback": str(tb),
            "error_file_name": error_file_name,
            "error_line_number": error_line_number,
            "error": e.__str__(),
        }

        
        exit_summary = OrderedDict(e_sum)

        
        yaml.add_representer(OrderedDict, util.represent_yaml_from_collections_dot_OrderedDict)
        

        try:
            jsonschema.validate(instance=exit_summary, schema=config.exit_summary_schema)
        except jsonschema.ValidationError as e:
            sys.stderr.write("Failed to validate the exit summary. Internal Error.\n")
            raise e
                
        self.print_github_block()

        """
        Print last n lines of log
        """



        
        for i in range(N):

            try:
                if self._loggable:
                    sys.stderr.write("{0} - last line of log\n".format(n_logs - i))
                    self.logger.log_it(logs[i], "ERROR")
                else:
                    sys.stderr.write("{0} - last line of log\n".format(n_logs - i))
                    sys.stderr.write(logs[i])
            except Exception as e:
                raise e
                
        sys.stderr.write("-" * 80 + "\n")
        if self._loggable:
            
            self.logger.log_it("...displaying last {0} lines of the log. Please see '{1}' for more details...".format(n_logs, self.logfile), "ERROR")
            self.logger.log_it(e.__str__())

            sys.stderr.write(THREE_LINES)

            sys.stderr.write(DNA_SPACER_1)

            sys.stderr.write(THREE_LINES)

            self.logger.log_it("="*40 + "\n", "ERROR")

            self.logger.log_it(" "*20 + "ERROR: Program exit summary:\n", "ERROR")
            
            self.logger.log_it("="*40 + "\n", "ERROR")
            
            self.logger.log_it("\n" + yaml.dump(exit_summary), "ERROR")

            self.logger.log_it("="*40 + "\n")

            sys.stderr.write(THREE_LINES)
            
            
        else:
            sys.stderr.write("...displaying last {0} lines of the log. Please see '{1}' for more details...\n".format(n_logs, self.logfile))
            sys.stderr.write(e.__str__())

            sys.stderr.write(THREE_LINES)

            sys.stderr.write(DNA_SPACER_1)

            sys.stderr.write(THREE_LINES)

            sys.stderr.write("="*40 + "\n\n")

            sys.stderr.write(" "*20 + "ERROR: Program exit summary:\n\n")
            
            sys.stderr.write("="*40 + "\n")
            
            sys.stderr.write("\n" + yaml.dump(exit_summary) + "\n")

            sys.stderr.write("="*40 + "\n")

            sys.stderr.write(THREE_LINES)



            
        return exit_summary


    











'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''



import io
import sys
import os
import gzip
import tempfile
import yaml, json
from collections import deque, OrderedDict
import psutil
import numpy as np
import math
import re

#import pdb

from builtins import open as _open

import jsonschema
from Bio import SeqIO, bgzf

#sys.path.append('..')

from kmerdb import kmer, util, config, appmap

# Logging configuration
global logger
logger = None

is_integer = re.compile("^[-+]?[0-9]+$")

def open(filepath, mode="r", metadata=None, sort:bool=False, slurp:bool=False, with_index:bool=False, logger=None):
    """
    Opens a file for reading or writing. Valid modes are 'xrwbt'. 'metadata=' is needed when writing/creating.
    Returns a lazy-loading KDBReader object or a KDBWriter object.
    The data may be force loaded with 'slurp=True'

    :param filepath:
    :type filepath: str
    :param mode:
    :type mode: str
    :param metadata: The file header/metadata dictionary to write to the file.
    :type metadata: dict
    :param sort: Sort on read the data into KDBReader
    :type sort: bool
    :param slurp: Immediately load all data into KDBReader
    :type slurp: bool
    :param with_index: Create index data dynamically.
    :type with_index: bool
    :returns: kmerdb.fileutil.KDBReader/kmerdb.fileutil.KDBWriter
    :rtype: kmerdb.fileutil.KDBReader
    """
    if type(filepath) is not str:
        raise TypeError("kmerdb.fileutil.open expects a str as its first positional argument")
    ## We don't want this because it won't write a new .kdb file on mode='w'
    # elif os.access(filepath, os.R_OK) is False:
    #     raise ValueError("kmerdb.fileutil.open expects a valid filepath as its first positional argument")
    elif type(mode) is not str:
        raise TypeError("kmerdb.fileutil.open expects the keyword argument 'mode' to be a str")

    if (mode == "w" or mode == "x") and (metadata is not None and (isinstance(metadata, OrderedDict) or type(metadata) is dict)):
        pass
    elif (mode == "w" or mode == "x"):
        raise TypeError("kmerdb.fileutil.open expects an additional metadata dictionary")
    elif sort is None or type(sort) is not bool:
        raise TypeError("kmerdb.fileutil.open expects a boolean for the keyword argument 'sort'")
    elif slurp is None or type(slurp) is not bool:
        raise TypeError("kmerdb.fileutil.open expects a boolean for the keyword argument 'slurp'")
    elif with_index is None or type(with_index) is not bool:
        raise TypeError("kmerdb.fileutil.open expects a boolean for the keyword argument 'with_index'")

    
    modes = set(mode)
    if modes - set("xrwbt") or len(mode) > len(modes):
        raise ValueError("invalid mode: {}".format(mode))


    
    creating = "x" in modes
    reading  = "r" in modes
    writing  = "w" in modes
    binary   = "b" in modes
    text     = "t" in modes

    if text and binary:
        raise ValueError("can't have text and binary mode at once")
    elif not (creating or reading or writing):
        raise ValueError("must have exactly one or read/write")

    if "r" in mode.lower():
        return KDBReader(filename=filepath, mode=mode, sort=sort, slurp=slurp, with_index=with_index, logger=logger)
    elif "w" in mode.lower() or "x" in mode.lower():
        return KDBWriter(metadata, filename=filepath, mode=mode, logger=logger)
    else:
        raise ValueError("Bad mode %r" % mode)





class KDBReader(bgzf.BgzfReader):
    """
    A class that reads .kdb files, potentially just for accessing header metadata, or for reading the entire contents into numpy arrays.
    

    :ivar filename: str
    :ivar fileobj: io.IOBase
    :ivar mode: str
    :ivar max_cache: int
    :ivar column_dtype: NumPy uint datatype
    :ivar count_dtypes: Numpy uint datatype
    :ivar frequencies_dtype: NumPy float datatype
    :ivar sort: bool
    :ivar slurp: bool
    :ivar with_index: bool
    """
    def __init__(self, filename:str=None, fileobj:io.IOBase=None, mode:str="r", max_cache:int=100, column_dtype:str="uint64", count_dtype:str="uint64", frequencies_dtype:str="float64", sort:bool=False, slurp:bool=False, with_index:bool=False, logger=None):
        if fileobj is not None and not isinstance(fileobj, io.IOBase):
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'fileobj' to be a file object")
        elif filename is not None and type(filename) is not str:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'filename' to be a str")
        elif os.access(filename, os.R_OK) is False:
            raise ValueError("kmerdb.fileutil.KDBReader expects a valid filepath as its first positional argument")

        elif sort is None or type(sort) is not bool:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'sort' to be a bool")
        elif max_cache is None or type(max_cache) is not int:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'max_cache' to be a int")
        elif mode is None or type(mode) is not str:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'mode' to be a str")
        elif column_dtype is None or type(column_dtype) is not str:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'column_dtype' to be a str")
        elif count_dtype is None or type(count_dtype) is not str:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'count_dtype' to be a str")
        elif frequencies_dtype is None or type(frequencies_dtype) is not str:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'frequencies_dtype' to be a str")
        elif slurp is None or type(slurp) is not bool:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'slurp' to be a bool")
        elif with_index is None or type(with_index) is not bool:
            raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'with_index' to be a bool")
        # elif logger is None:
        #     raise TypeError("kmerdb.fileutil.KDBReader expects the keyword argument 'logger' to be valid")

        
        
        
        if fileobj:
            assert filename is None
            handle = fileobj
            assert "b" in handle.mode.lower()
        else:
            if "w" in mode.lower() or "a" in mode.lower():
                raise ValueError("Must use read mode (default), not write or append mode")
            handle = _open(filename, "rb")

        self._text = "b" not in mode.lower()
        if self._text:
            self._newline = "\n"
        else:
            self._newline = b"\n"
        self._handle      = handle
        self._filepath    = self._handle.name
        self.max_cache    = max_cache
        self._buffers     = {}
        self._block_start_offset = None
        self._block_raw_length = None
        self.kmer_ids     = None
        self.counts       = None
        self.frequencies  = None
        self.count_dtype = None
        self.frequencies_dtype = None
        self.kmer_ids_dtype = None


        # Has been slurped
        self.completed = False


        
        self.logger = logger

        self._loggable = logger is not None
        
        '''
        Here we want to load the metadata blocks. We want to load the first two lines of the file: the first line is the version, followed by the number of metadata blocks
        '''
        # 0th block
        if self._loggable:
            self.logger.log_it("Loading the 0th block from '{0}'...".format(self._filepath), "INFO")
        self._load_block(self._handle.tell())

        self._buffer = self._buffer.rstrip(config.header_delimiter)
        
        initial_header_data = OrderedDict(yaml.safe_load(self._buffer))
        num_header_blocks = None

        if type(initial_header_data) is str:
            if self._loggable:
                self.logger.log_it("Inappropriate type for the header data.", "INFO")
            #logger.info("Um, what the fuck is this in my metadata block?")
            raise TypeError("kmerdb.fileutil.KDBReader could not parse the YAML formatted metadata in the first blocks of the file")
        elif type(initial_header_data) is OrderedDict:
            if self._loggable:
                self.logger.log_it("Successfully parsed the 0th block of the file, which is expected to be the first block of YAML formatted metadata", "INFO")
                self.logger.log_it("Assuming YAML blocks until delimiter reached.", "INFO")
            if "version" not in initial_header_data.keys():
                raise TypeError("kmerdb.fileutil.KDBReader couldn't validate the header YAML")
            elif "metadata_blocks" not in initial_header_data.keys():
                raise TypeError("kmerdb.fileutil.KDBReader couldn't validate the header YAML")
            else:
                if self._loggable:
                    self.logger.log_it(str(initial_header_data), "INFO")
                if initial_header_data["metadata_blocks"] == 1:
                    if self._loggable:
                        self.logger.log_it("1 metadata block: Not loading any additional blocks", "INFO")
                    pass
                else:
                    for i in range(initial_header_data["metadata_blocks"] - 1):
                        self._load_block(self._handle.tell())
                        if self._loggable:
                            self.logger.log_it("Multiple metadata blocks read...", "DEBUG")
                            self.logger.log_it("Secondary blocks read", "DEBUG")
                        addtl_header_data = yaml.safe_load(self._buffer.rstrip(config.header_delimiter))
                        if type(addtl_header_data) is str:
                            if self._loggable:
                                self.logger.log_it(addtl_header_data, "ERROR")
                                self.logger.log_it("Couldn't determine this block.::/", "ERROR")
                            raise TypeError("kmerdb.fileutil.KDBReader determined the data in the {0} block of the header data from '{1}' was not YAML formatted".format(i, self._filepath))
                        elif type(addtl_header_data) is dict:
                            sys.stderr.write("\r")
                            sys.stderr.write("Successfully parsed {0} blocks of YAML formatted metadata".format(i))
                            initial_header_data.update(addtl_header_data)
                            num_header_blocks = i
                        else:
                            if self._loggable:
                                self.logger.log_it(addtl_header_data, "ERROR")
                            raise RuntimeError("kmerdb.fileutil.KDBReader encountered a addtl_header_data type that wasn't expected when parsing the {0} block from the .kdb file '{1}'.".format(i, self._filepath))
        else:
            raise RuntimeError("kmerdb.fileutil.KDBReader encountered an unexpected type for the header_dict read from the .kdb header blocks")
        if self._loggable:
            self.logger.log_it("Reading additional blocks as YAML...", "INFO")
            self.logger.log_it("Validating the header data against the schema...", "INFO")
        try:
            jsonschema.validate(instance=initial_header_data, schema=config.kdb_metadata_schema)
            self.metadata = dict(initial_header_data)



            # Metadata values
            self.k = self.metadata['k']
            self.kmer_ids_dtype = self.metadata["kmer_ids_dtype"]
            self.count_dtype = self.metadata["count_dtype"]
            self.frequencies_dtype = self.metadata["frequencies_dtype"]
            self.sorted = self.metadata["sorted"]


            # Logging statements
            if self._loggable:
                self.logger.log_it("Self assigning dtype to uint64 probably", "DEBUG")
                self.logger.log_it("Checking for metadata inference...", "DEBUG")



            # .kdb id and count vectors
            self.kmer_ids = np.zeros(4**self.metadata["k"], dtype=self.metadata["kmer_ids_dtype"])
            self.counts = np.zeros(4**self.metadata["k"], dtype=self.metadata["count_dtype"])
            self.frequencies = np.zeros(4**self.metadata["k"], dtype=self.metadata["frequencies_dtype"])



            # .kdb index
            self.index = None
            self.with_index = with_index
            
        except jsonschema.ValidationError as e:
            if self._loggable:
                self.logger.log_it("kmerdb.fileutil.KDBReader couldn't validate the header/metadata YAML from {0} header blocks".format(num_header_blocks), "ERROR")
                self.logger.log_it("""


Failed to validate YAML header.

-------------------------------



 You can store unique k-mer counts, total nullomer counts, and other metadata alongside a k-mer count vector with the 'kmerdb profile' command



{0}

 ...then try again to view the header with 'kmerdb header' or the whole file : 'kmerdb view -H kmer_count_vector.8.kdb'



""".format(appmap.command_1_usage, "ERROR"))

                self.logger.log_it(e.__str__, "ERROR")
            raise ValueError("Requires kmerdb v{0} format YAML header. Body is .tsv format table, .bgzf file.      - k-mer count vector        (idx, k-mer id, count, frequency)".format(config.VERSION))

            


        self.metadata["header_offset"] = self._handle.tell()

        if self._loggable:
            self.logger.log_it("Handle set to {0} after reading header, saving as handle offset".format(self.metadata["header_offset"]), "DEBUG")
        #self._reader = gzip.open(self._filepath, 'r')
        self._offsets = deque()
        for values in bgzf.BgzfBlocks(self._handle):
            #logger.debug("Raw start %i, raw length %i, data start %i, data length %i" % values)
            self._offsets.appendleft(values) # raw start, raw length, data start, data length
        if len(self._offsets) == 0:
            raise IOError("kmerdb.fileutil.KDBReader opened an empty file")
        # Skip the zeroth block
        self._load_block()
        # print(str(self._buffer)) # 1
        # print(self.readline())
        # self._load_block()
        # print(self._buffer) # 2
        # print(self.readline())

        if sort is True and self.metadata["sorted"] is True:
            sort is True
        else:
            sort is False

        if slurp is True:
            self.slurp(column_dtypes=column_dtype, count_dtypes=count_dtype, frequencies_dtype=frequencies_dtype, sort=sort, with_index=with_index)
            

        self.kmer_ids_dtype = column_dtype
        self.count_dtypes = count_dtype
        self.frequencies_dtype = frequencies_dtype
        handle.close()
        self._handle.close()

        if slurp is True:
            self._handle = None
            handle = None
        fileobj=None
        return

        #
    
    def __iter__(self):
        return self

    def __exit__(self, type, value, tb):
        if self._handle is not None:
            self._handle.close()
        return



    

    def _slurp(self, column_dtypes:str="uint64", count_dtypes:str="uint64", frequencies_dtype:str="float64", sort:bool=False, with_index:bool=False):
        if type(column_dtypes) is not str:
            raise TypeError("kmerdb.fileutil.KDBReader.slurp expects the column_dtypes keyword argument to be a str")
        elif type(count_dtypes) is not str:
            raise TypeError("kmerdb.fileutil.KDBReader.slurp expects the count_dtypes keyword argument to be a str")
        elif type(frequencies_dtype) is not str:
            raise TypeError("kmerdb.fileutil.KDBReader.slurp expects the frequencies_dtype keyword argument to be a str")

        try:
            np.dtype(column_dtypes)
            np.dtype(count_dtypes)
            np.dtype(frequencies_dtype)
        except TypeError as e:
            if self._loggable:
                self.logger.log_it(e.__str__(), "ERROR")
                self.logger.log_it("kmerdb.fileutil.KDBReader.slurp encountered a TypeError while assessing a numpy dtype", "ERROR")
            raise TypeError("kmerdb.fileutil.KDBReader._slurp expects the dtype keyword argument to be a valid numpy data type")
        if sort is None or type(sort) is not bool:
            raise TypeError("kmerdb.fileutil.KDBReader._slurp expects the 'sort' keyword argument to be a bool")
        elif with_index is None or type(with_index) is not bool:
            raise TypeError("kmerdb.fileutil.KDBReader._slurp expects the 'with_index' keyword argument to be a bool")
        if self._handle is None:
            self._handle = _open(self._filepath, 'rb')
        # First calculate the amount of memory required by the array
        N = 4**self.k # The dimension of the k-space, or the number of elements for the array
        num_bytes = 4 * N
        #logger.info("Approximately {0} bytes".format(num_bytes))
        vmem = psutil.virtual_memory()


        # Initial zero-valued empty id/count vectors
        self.kmer_ids = np.zeros(N, dtype=column_dtypes)
        self.counts = np.zeros(N, dtype=count_dtypes)
        self.frequencies = np.zeros(N, dtype=frequencies_dtype)

        
        self.all_kmer_metadata = []


        # Initial Python lists
        kmer_ids = []
        profile = []
        counts = []
        frequencies = []


        # Block offsets
        offsets = []

        
        if vmem.available > num_bytes:
            # Do the slurp
            if self._loggable:
                self.logger.log_it("Reading profile into memory", "INFO")
            if sort is False:
                i = 0
                try:
                    for j in range(N):
                        #logger.debug("Reading {0}th line...".format(j))
                        try:
                            line = next(self)
                            offset = self._handle.tell()
                        except StopIteration as e:
                            if self._loggable:
                                self.logger.log_it("Finished loading initial profile through slurp-on-init", "ERROR")
                                self.logger.log_it("Read profile from: {0}".format(self._filepath), "ERROR")
                            #raise e
                            pass
                        if line is None:
                            if self._loggable:
                                self.logger.log_it("'next' was None... Internal Error", "ERROR")
                            raise RuntimeError("Could not complete parsing of file. Internal Error")
                        # Don't forget to not parse the metadata column [:-1]

                        l = line.rstrip().split("\t")
                        try:
                            assert len(l) == config.KDB_COLUMN_NUMBER, "kmerdb.fileutil.KDBReader requires .kdb files to have {0} columns.".format(config.KDB_COLUMN_NUMBER)
                        except AssertionError as e:
                            self.logger.log_it("Found .kdb table line with {0} columns. Requires {1} columns.".format(l, config.KDB_COLUMN_NUMBER))
                            self.logger.log_it(line, "ERROR")
                            raise e
                            
                        x, kmer_id, _count, _frequency = l
                        x = int(x)
                        kmer_id = int(kmer_id)                            

                        assert j == x, "Line number did not match"
                        
                        kmer_ids.append(kmer_id)

                        count = int(_count)
                        frequency = float(count)/N
                        # K-mer id and count NumPy index-based assignment
                        self.kmer_ids[j] = kmer_id
                        self.counts[kmer_id] = count
                        self.frequencies[kmer_id] = frequency


                        # Block offsets
                        if with_index is True:
                            offsets.append((kmer_id, offset))
                    
                        i += 1

                    
                except StopIteration as e:
                    if i == N:
                        raise e
                    else:
                        if self._loggable:
                            self.logger.log_it("Number of lines read: {0}".format(i), "DEBUG")
                            self.logger.log_it("Number of k-mers (N) set globally.".format(N), "DEBUG")
                            self.logger.log_it("Read only {0} lines from the file...".format(i), "DEBUG")
                            self.logger.log_it("Profile must have been read before this point", "DEBUG")
                            self.logger.log_it(e.__str__(), "ERROR")
                        raise e
                except Exception as e:
                    raise e
                assert self.counts.size == self.kmer_ids.size, "Counts size has diverged from the kmer_ids shape"
                assert self.counts.size == self.frequencies.size, "Frequencies size has diverged from counts shape"
            elif sort is True:
                i = 0
                try:
                    for j in range(N):
                        #logger.debug("Reading {0}th line...".format(j))
                        try:
                            line = next(self)
                            offset = self._handle.tell()
                        except StopIteration as e:
                            self.logger.log_it("Finished loading initial profile through slurp-on-init", "ERROR")
                            raise e
                        if line is None:
                            self.logger.log_it("Next was None... profile was sparse, breaking", "WARNING")
                            raise IOError("next method was None on sorted import. Internal Error.")

                        # Don't forget to not parse the metadata column [:-1]

                        x, kmer_id, _count, _frequency = line.rstrip().split("\t")


                        _frequency = float(_frequency)
                        kmer_id = int(kmer_id)

                        s = is_integer.match(_count)
                        if s is not None:
                            count = int(_count)
                        else:
                            raise TypeError("kmerdb.fileutil.KDBReader._slurp | couldn't properly detect integer")

                        frequency = float(count)/N
                        # Trying to get the needle threaded here.
                        # The goal is for the profile to be a running index, similar to the row number
                        # But this could maintain lexical sort order
                        # 
                        self.kmer_ids[j] = kmer_id
                        self.counts[kmer_id] = count
                        self.frequencies[kmer_id] = _frequency


                        # Block offsets
                        if with_index is True:
                            offsets.append((kmer_id, offset))


                        # The index data, just a list of id and gzip file-offsets
                        i+=1
                        #sys.stderr.write("::DEBUG::   |\-)(||||..... KMER_ID: {0} COUNT: {1}".format(kmer_id, count))
                    assert self.kmer_ids.size == self.counts.size, "Counts size has diverged from kmer_ids size/shape"
                    assert self.counts.size == self.frequencies.size, "Frequencies size has diverged from counts shape"

                    # Index tuple  data
                    #offsets.append((kmer_id, offset))


                    
                except StopIteration as e:
                    if i == N:
                        raise e
                    else:
                        raise e
                except Exception as e:
                    raise e
                if self._loggable:
                    self.logger.log_it("Read {0} lines from the file...".format(i), "INFO")
                assert self.kmer_ids.size == self.counts.size, "Counts size is mismatched in with kmer_ids size"
                assert self.frequencies.size == self.counts.size, "Frequencies size is mismatched in count from counts size"
                if self._loggable:
                    self.logger.log_it("Correct array sizes match...", "DEBUG")
                self._handle.seek(0)
                self._load_block()
                #logger.debug("Dammit, why can't i reset the Bio.bgzf filehandle...")
                if sort is True:
                    # If the file is sorted, do not sort
                    kmer_ids_sorted_by_count = np.lexsort((self.kmer_ids, self.counts))
                    reverse_kmer_ids_sorted_by_count = np.flipud(kmer_ids_sorted_by_count)
                    for i, idx in enumerate(kmer_ids_sorted_by_count): # This is right, not fixing this.
                        kmer_id = self.kmer_ids[idx]
                        count = self.counts[idx]

                        ## I stand corrected. m
                        #self.kmer_ids[i] = kmer_ids[profile[i]]
                        #self.counts[idx] = counts[idx]
                        ###self.frequencies[idx] = frequencies[idx]
                        #logger.debug("Just in casey eggs and bakey...")

            else:
                raise RuntimeError("Internal error. Failed to initialize array when determining sort strategy")
        else:
            raise OSError("The dimensionality at k={0} or 4^k = {1} exceeds the available amount of available memory (bytes) {2}".format(self.k, N, vmem.available))



        # print(offsets)
        # print("Offsets printed...")
        # raise RuntimeError("yikes")

        if with_index is True:
            self.index = offsets



        
        if self.kmer_ids.dtype != self.kmer_ids_dtype:
            raise TypeError("kmerdb.fileutil.KDBReader._slurp encountered an awful TypeError")
        elif self.counts.dtype != self.count_dtype:
            raise TypeError("kmerdb.fileutil.KDBReader._slurp encountered an awful TypeError")
        elif self.frequencies.dtype != self.frequencies_dtype:
            raise TypeError("kmerdb.fileutil.KDBReader._slurp encountered an awful TypeError")
        #self.kmer_ids = np.array(kmer_ids, dtype=suggested_dtype)
        #self.kmer_ids = np.array(kmer_ids, dtype=column_dtypes)
        #self.counts = np.array(counts, dtype=count_dtypes)
        #self.frequencies = np.array(frequencies, dtype=frequencies_dtype)
        self._handle.seek(0)
        self._load_block()
        #print([x for x in np.ndindex(self.kmer_ids.flat) if x < ])

        assert self.kmer_ids.size == self.counts.size, "Number of Counts mismatched in size from kmer_id array size"
        assert self.frequencies.size == self.counts.size, "Number of Frequencies is mismatched in size from Count array size"

        if with_index is True:
            assert len(self.index) == self.kmer_ids.size, "Number of index elements is mismatched in size from kmer_id array size"


        
        self.completed = True
        return self.counts

    def slurp(self, column_dtypes:str="uint64", count_dtypes:str="uint64", frequencies_dtype:str="float64", sort:bool=False, with_index:bool=False):
        """
        A function to lazy-load an entire .kdb file into memory. 

        :param column_dtypes: a NumPy uint datatype
        :type column_dtypes: str
        :param count_dtypes: a NumPy uint datatype
        :type count_dtypes: str
        :param frequencies_dtype: a NumPy float datatype
        :type frequencies_dtype: str
        :param sort: Whether or not to sort the columns?
        :type sort: bool
        :param with_index: dynamically create index offsets while loading
        :type with_index: bool
        """

        if np.sum(self.counts) != 0:
            return self.counts
        else:
            return self._slurp(column_dtypes=column_dtypes, count_dtypes=count_dtypes, frequencies_dtype=frequencies_dtype, sort=sort, with_index=with_index)
        


    
    
class KDBWriter(bgzf.BgzfWriter):
    """
    A wrapper class around Bio.bgzf.BgzfWriter to write a .kdb file to disk.

    :ivar metadata: OrderedDict
    :ivar filename: str
    :ivar mode: str
    :ivar fileobj: io.IOBase
    :ivar compresslevel: int
    """
    
    def __init__(self, metadata:OrderedDict, filename=None, mode="w", fileobj=None, compresslevel=6, logger=None):
        """Initilize the class."""

        self.logger = logger
        self._loggable = logger is not None


        if type(metadata) is not dict and not isinstance(metadata, OrderedDict):
            raise TypeError("kmerdb.fileutil.KDBWriter expects a valid metadata dictionary as its first positional argument")
        try:
            if self._loggable:
                self.logger.log_it("Validating metadata schema against the config.py header schema", "DEBUG")
            jsonschema.validate(instance=dict(metadata), schema=config.kdb_metadata_schema)
            self.metadata = metadata
            self.k = self.metadata['k']
        except jsonschema.ValidationError as e:
            if self._loggable:
                self.logger.log_it("kmerdb.fileutil.KDBReader couldn't validate the header YAML", "ERROR")
            raise e

        if fileobj:
            assert filename is None
            handle = fileobj
        else:
            if "w" not in mode.lower() and "a" not in mode.lower():
                raise ValueError("Must use write or append mode, not %r" % mode)
            if "a" in mode.lower():
                raise NotImplementedError("Append mode is not implemented yet")
                # handle = _open(filename, "ab")
            else:
                handle = _open(filename, "wb")
        self._text = "b" not in mode.lower()
        self._handle = handle
        self._buffer = b"" if "b" in mode.lower() else ""
        self.compresslevel = compresslevel


        
        """
        Write the header to the file
        """



        if self._loggable:
            self.logger.log_it("Constructing a new .kdb file '{0}'...".format(self._handle.name), "INFO")
        yaml.add_representer(OrderedDict, util.represent_yaml_from_collections_dot_OrderedDict)
        if "b" in mode.lower():
            metadata_bytes = bytes(yaml.dump(self.metadata, sort_keys=False), 'utf-8')
            metadata_plus_delimiter_in_bytes = metadata_bytes + bytes(config.header_delimiter, 'utf-8')
            self.metadata["metadata_blocks"] = math.ceil( sys.getsizeof(metadata_plus_delimiter_in_bytes) / ( 2**16 ) ) # First estimate
            metadata_bytes = bytes(yaml.dump(self.metadata, sort_keys=False), 'utf-8')
            metadata_bytes = metadata_bytes + bytes(config.header_delimiter, 'utf-8')
            self.metadata["metadata_blocks"] = math.ceil( sys.getsizeof(metadata_bytes) / ( 2**16 ) ) # Second estimate
            metadata_bytes = bytes(yaml.dump(self.metadata, sort_keys=False), 'utf-8')
            metadata_bytes = metadata_bytes + bytes(config.header_delimiter, 'utf-8')
            if self._loggable:
                self.logger.log_it("Writing the {0} metadata blocks to the new file".format(self.metadata["metadata_blocks"]), "INFO")

            # 01-01-2022 This is still not a completely functional method to write data to bgzf through the Bio.bgzf.BgzfWriter class included in BioPython
            # 04-10-2023 This is still disgusting to me. I understand a limited amount of the Bio.bgzf source or the nuances of the format specification
            # That said, it's producing files, with data in the correct order. The metadata_blocks/offset calculation is still rudimentary
            # But I'm able to generate my file format reasonably.
            # I've needed to implement a basic block_writer, maintaining compatibility with the Biopython bgzf submodule.
            #self.write(bytes(yaml.dump(metadata, sort_keys=False), 'utf-8'))
        
            for i in range(self.metadata["metadata_blocks"]):
                metadata_slice = metadata_bytes[:65536]
                metadata_bytes = metadata_bytes[65536:]
                self._write_block(metadata_slice)

                #self._write_block
                self._buffer = b""
                self._handle.flush()
        elif "w" == mode.lower() or "x" == mode.lower():
            metadata_yaml_str = yaml.dump(metadata, sort_keys=False)
            metadata_bytes = bytes(metadata_yaml_str, 'utf-8')
            mode = "b"
            self._buffer = b""            
            self.write(metadata_bytes)

            self._handle.flush()
        else:
            raise RuntimeError("Could not determine proper encoding for write operations to .kdb file")


class FileReader:
    def __init__(self, args, logger=None):
        self.arguments = args
        self.logger = logger

        
    def parsefile(self, filename):
        """Wrapper function for the KDBReader to keep arguments succinct for deployment through multiprocessing.Pool
            
        :param filename: the filepath of the fasta(.gz)/fastq(.gz) to process with parsefile -> parse.SeqParser
        :type filename: str
        :returns: (db, m, n)
        """
        return open(filename, mode='r', slurp=True, logger=self.logger)
        



import sys
import os
import unittest
import yaml


from kmerdb import fileutil



sys.path.append("..")

filepath_that_doesnt_exist = os.path.join(os.path.dirname(__file__), "data/foo.bar")
valid_kdb_filepath = os.path.join(os.path.dirname(__file__), "data/test_Cac_ATCC824.8.kdb")
valid_kdb_metadata_str = """
count_dtype: uint64
files:
- filename: test/data/Cacetobutylicum_ATCC824.fasta.gz
  md5: 0a0f73e1c8b8285703e29279bafaabef
  nullomers: 1433
  sha256: f9081291b62ff3387f1ca6ee2484669c849ed1840fdf2dd9dc3a0c93e9e87951
  total_kmers: 4132866
  total_reads: 2
  unique_kmers: 64103
frequencies_dtype: float64
header_offset: 324
k: 8
kmer_ids_dtype: uint64
metadata: false
metadata_blocks: 1
profile_dtype: uint64
sorted: false
tags: []
total_kmers: 4132866
unique_kmers: 64103
unique_nullomers: 1433
version: 0.8.3
"""
 


class TestDoesInputFileExist(unittest.TestCase):
    def given_a_filepath_that_does_not_exist(self):
        self.filepath_that_doesnt_exist = "data/foo.bar"
        
        
    def then_fileutil_KDBReader_should_throw_a_ValueError(self):

        self.assertRaises(ValueError, fileutil.open, self.filepath_that_doesnt_exist, mode='r', slurp=False)

    def test_nonexistent_kdb_file_with_fileutil_KDBReader(self):
        """
        Actual UnitTest test
        """
        self.given_a_filepath_that_does_not_exist()
        self.then_fileutil_KDBReader_should_throw_a_ValueError()
        



class TestValidKDBFile(unittest.TestCase):


    def given_a_valid_kdb_file(self):
        """ Given a valid .kdb file """
        self.test_kdb_metadata_str = valid_kdb_metadata_str
        self.test_kdb_metadata = yaml.safe_load(self.test_kdb_metadata_str)
        self.test_kdb_filepath = valid_kdb_filepath
        
    def when_checking_if_kdb_file_is_valid(self):
        self.test_valid_kdb = fileutil.open(os.path.join(os.path.dirname(__file__), self.test_kdb_filepath), mode='r', slurp=True)

    def then_yaml_metadata_header_should_be_valid(self):

        
        self.assertTrue(self.test_kdb_metadata == self.test_valid_kdb.metadata)
        self.assertTrue(self.test_valid_kdb.k == 8)

    def then_kmerid_and_count_array_dimensions_should_be_4_power_8(self):
        self.assertTrue(len(self.test_valid_kdb.kmer_ids) == 65536)
        self.assertTrue(len(self.test_valid_kdb.counts) == 65536)

    def test_valid_kdb_file_with_fileutil_KDBReader(self):
        """
        Actual UnitTest test
        """
        self.given_a_valid_kdb_file()
        self.when_checking_if_kdb_file_is_valid()
        self.then_yaml_metadata_header_should_be_valid()
        self.then_kmerid_and_count_array_dimensions_should_be_4_power_8()

    def given_a_filepath_to_an_invalid_kdb_file(self):
        self.filepath_to_invalid_kdb_file = os.path.join(os.path.dirname(__file__), "data/test_invalid_kdb_file.kdb")


    def then_fileutil_open_should_throw_a_ValueError(self):
        # with fileutil.open(self.invalid_kdb_filepath, 'r', slurp=False) as fh:
        #     self.assertRaises(ValueError)
        self.assertRaises(ValueError, fileutil.open, self.filepath_to_invalid_kdb_file, mode='r', slurp=False)
        

    def test_invalid_kdb_file_with_fileutil_open(self):
        """
        Actual UnitTest test
        """

        self.given_a_filepath_to_an_invalid_kdb_file()
        self.then_fileutil_open_should_throw_a_ValueError()



class TestErrorsRaisedByOpen(unittest.TestCase):


    def given_a_filepath_that_is_not_a_string(self):
        self.nonstring_filepath = False
        
    def then_open_raises_TypeError_from_invalid_filepath(self):
        self.assertRaises(TypeError, fileutil.open, self.nonstring_filepath, mode='r', slurp=False)

    def test_invalid_filepath_type(self):
        self.given_a_filepath_that_is_not_a_string()
        self.then_open_raises_TypeError_from_invalid_filepath()

        
    def given_a_mode_that_is_not_a_string(self):
        self.nonstring_mode = False
        self.valid_kdb_filepath = valid_kdb_filepath
        
    def then_open_raises_TypeError_from_invalid_mode_type(self):
        self.assertRaises(TypeError, fileutil.open, self.valid_kdb_filepath, mode=self.nonstring_mode, slurp=False)

    def test_invalid_mode_string_type(self):
        self.given_a_mode_that_is_not_a_string()
        self.then_open_raises_TypeError_from_invalid_mode_type()
        

    def given_a_mode_that_is_not_recognized(self):
        self.invalid_mode_string = 'z'
        self.valid_kdb_filepath = valid_kdb_filepath

    def then_open_raises_ValueError_from_invalid_mode_string(self):
        self.assertRaises(ValueError, fileutil.open, self.valid_kdb_filepath, mode=self.invalid_mode_string, slurp=False)

    def test_invalid_mode_string(self):

        self.given_a_mode_that_is_not_recognized()
        self.then_open_raises_ValueError_from_invalid_mode_string()

    def given_a_slurp_that_is_not_a_bool(self):
        self.invalid_slurp = None
        self.valid_kdb_filepath = valid_kdb_filepath

    def then_open_raises_TypeError_from_invalid_slurp(self):
        self.assertRaises(TypeError, fileutil.open, self.valid_kdb_filepath, mode='r', slurp=self.invalid_slurp)

    def test_invalid_slurp_type(self):
        self.given_a_slurp_that_is_not_a_bool()

        self.then_open_raises_TypeError_from_invalid_slurp()
        

class TestOpenReturnsReaderOrWriterBasedOnMode(unittest.TestCase):

    def given_a_mode_string_to_read_from_filepath(self):
        self.valid_kdb_filepath = valid_kdb_filepath
        self.read_mode = 'r'

    def then_open_returns_KDBReader(self):
        self.assertTrue(isinstance(fileutil.open(self.valid_kdb_filepath, mode=self.read_mode, slurp=False), fileutil.KDBReader))

    def test_open_returns_KDBReader_on_read_mode(self):
        self.given_a_mode_string_to_read_from_filepath()
        self.then_open_returns_KDBReader()
        
    def given_a_mode_string_to_write_to_filepath(self):
        self.valid_kdb_filepath = valid_kdb_filepath
        self.write_mode = 'w'
        self.filepath_that_doesnt_exist = filepath_that_doesnt_exist
        self.test_kdb_metadata_str = valid_kdb_metadata_str
        self.test_kdb_metadata = yaml.safe_load(self.test_kdb_metadata_str)

        
    def then_open_returns_KDBWriter(self):
        self.assertTrue(isinstance(fileutil.open(filepath_that_doesnt_exist, mode=self.write_mode, metadata=self.test_kdb_metadata), fileutil.KDBWriter))

    def test_open_returns_KDBWriter_on_write_mode(self):
        self.given_a_mode_string_to_write_to_filepath()
        self.then_open_returns_KDBWriter()

    def tearDown(self):
        if os.access(filepath_that_doesnt_exist, os.R_OK):
            os.unlink(filepath_that_doesnt_exist)
        

    
if __name__ == "__main__":
    unittest.main()
    
        
